"""Expression exporting interface implementation.
"""

import os
import re
from logging import Logger
from typing import List
from typing import Match
from typing import Optional
from typing import Pattern

from html_minifier.minify import Minifier

from apysc._console import loggers

info_logger: Logger = loggers.get_info_logger()


def save_overall_html(
        dest_dir_path: str,
        *,
        html_file_name: str = 'index.html',
        minify: bool = True,
        js_lib_dir_path: str = './',
        skip_js_lib_exporting: bool = False,
        embed_js_libs: bool = False,
        verbose: int = 1) -> None:
    """
    Save the overall HTML and js files under the specified
    directory path.

    Notes
    -----
    This interface empties a specified directory before saving.

    Parameters
    ----------
    dest_dir_path : str
        Destination directory path to save each HTML and js file.
    html_file_name : str, default 'index.html'
        The output HTML file name.
    minify : bool, default True
        Boolean value indicates whether minify HTML and js or not.
        The False setting is helpful when debugging.
    js_lib_dir_path : str, default './'
        JavaScript libraries directory path. This setting
        applies to a JavaScript source path in HTML.
        If not specified, then set the same directory with HTML.
        This setting is maybe helpful to set js lib directory,
        such as Django's static (static_collected) directory.
        This interface recommends setting True value to the
        `skip_js_lib_exporting` argument if this argument sets.
    skip_js_lib_exporting : bool, default False
        If True, this interface does not export JavaScript libraries.
    embed_js_libs: bool, default False
        Option to embed the JavaScript libraries script to the
        output HTML or not. If True, the output HTML becomes enormous,
        and be only one HTML file. Occasionally, this option is
        useful when sharing the exported file or using the output
        file with an iframe tag to avoid the CORS error.
    verbose : int, default 1
        The Logging setting. If 0 is specified, this interface
        does not display a logging message. If 1 or the other
        value is specified, this interface displays a message usually.

    References
    ----------
    - save_overall_html interface document
        - https://simon-ritchie.github.io/apysc/save_overall_html.html

    Examples
    --------
    >>> import apysc as ap
    >>> stage: ap.Stage = ap.Stage()
    >>> # Do something here...
    >>> ap.save_overall_html(dest_dir_path='tmp/output/')
    """
    from apysc._file import file_util
    from apysc._html import html_util
    _display_debug_mode_ignoring_minify_setting_info(
        minify=minify, verbose=verbose)
    _display_info(msg='Overall exporting started...', verbose=verbose)
    file_util.empty_directory(directory_path=dest_dir_path)
    _display_info(msg='JavaScript libraries exporting...', verbose=verbose)
    _ = _export_js_libs(
        dest_dir_path=dest_dir_path,
        skip_js_lib_exporting=skip_js_lib_exporting)
    html_str: str = html_util.append_html_to_str(
        to_append_html='<html>', dest_html='', indent_num=0)
    html_str = _append_head_to_html_str(
        html_str=html_str, js_lib_dir_path=js_lib_dir_path,
        embed_js_libs=embed_js_libs)
    html_str = html_util.append_html_to_str(
        to_append_html='<body style="margin: 0;">',
        dest_html=html_str,
        indent_num=0)
    html_str = _append_stage_global_variable_to_html(html_str=html_str)
    html_str = _append_expression_to_html_str(
        html_str=html_str, verbose=verbose)
    html_str = html_util.append_html_to_str(
        to_append_html='</body>', dest_html=html_str, indent_num=0)
    html_str = _append_entry_point_function_call(html_str=html_str)
    html_str = html_util.append_html_to_str(
        to_append_html='</html>', dest_html=html_str, indent_num=0)
    html_str = _minify_html(html_str=html_str, minify=minify)
    _display_info(msg='HTML saving started...', verbose=verbose)
    _save_html(
        html_str=html_str, dir_path=dest_dir_path, file_name=html_file_name)
    file_path: str = os.path.join(dest_dir_path, html_file_name)
    _display_info(
        msg=f'All files were exported! \nFile path is : {file_path}',
        verbose=verbose)


def _display_debug_mode_ignoring_minify_setting_info(
        *, minify: bool, verbose: int) -> str:
    """
    Display information of ignoring minify setting if the
    debug mode is enabled.

    Parameters
    ----------
    minify : bool
        A Boolean value indicates whether minify HTML
        and js or not. The False setting is helpful
        when debugging.
    verbose : int
        If 0 is specified, the apysc does not display
        a message. The apysc displays a message when
        this value is 1 or the other value.

    Returns
    -------
    msg : str
        Displayed message.
    """
    import apysc as ap
    if not minify:
        return ''
    if not ap.is_debug_mode():
        return ''
    msg: str = (
        'The minify setting is ignored since the debug mode has been '
        'enabled.')
    _display_info(msg=msg, verbose=verbose)
    return msg


def _display_info(*, msg: str, verbose: int) -> str:
    """
    Display an info log message.

    Parameters
    ----------
    msg : str
        A message to display.
    verbose : int
        If 0 is specified, the apysc does not display
        a message. 1 or the other value displays the message.

    Returns
    -------
    msg : str
        Displayed message.
    """
    if verbose == 0:
        return ''
    info_logger.info(msg=msg)
    return msg


def _minify_html(*, html_str: str, minify: bool) -> str:
    """
    Minify HTML and js string.

    Notes
    -----
    If the debug mode setting is enabled, the apysc
    skips minifying.

    Parameters
    ----------
    html_str : str
        HTML string to minify.
    minify : bool
        A boolean value indicates whether minify HTML
        and js or not. If False, the apysc skips minifying.

    Returns
    -------
    html_str : str
        Result html string.
    """
    import apysc as ap
    if not minify:
        return html_str
    if ap.is_debug_mode():
        return html_str
    minifier = Minifier(html=html_str)
    html_str = minifier.minify()
    return html_str


def _append_stage_global_variable_to_html(*, html_str: str) -> str:
    """
    Append a stage's global variable to an HTML string.

    Parameters
    ----------
    html_str : str
        Target HTML string.

    Returns
    -------
    html_str : str
        After appended HTML string.
    """
    from apysc._display.stage import get_stage_elem_id
    from apysc._html import html_const
    from apysc._html import html_util
    html_str = html_util.append_html_to_str(
        to_append_html=html_const.SCRIPT_START_TAG,
        dest_html=html_str, indent_num=0)
    html_str = html_util.append_html_to_str(
        to_append_html=f'var {get_stage_elem_id()};',
        dest_html=html_str, indent_num=0)
    html_str = html_util.append_html_to_str(
        to_append_html=html_const.SCRIPT_END_TAG,
        dest_html=html_str, indent_num=0)
    return html_str


def get_entry_point_func_name() -> str:
    """
    Get an entry point function name.

    Returns
    -------
    entry_point_func_name : str
        An entry point function name.
    """
    from apysc._display.stage import get_stage_variable_name
    stage_variable_name: str = get_stage_variable_name()
    entry_point_func_name: str = f'main_{stage_variable_name}'
    return entry_point_func_name


def _append_entry_point_function_call(*, html_str: str) -> str:
    """
    Append entry-point function call's script to a HTML string.

    Parameters
    ----------
    html_str : str
        Target HTML string.

    Returns
    -------
    html_str : str
        After appended html string.
    """
    import apysc as ap
    html_str += (
        '\n<script type="text/javascript">'
        f'\n$({ap.document.variable_name}).ready(function() {{'
    )
    entry_point_func_name: str = get_entry_point_func_name()
    html_str += f'\n  {entry_point_func_name}();'
    html_str += (
        '\n});'
        '\n</script>'
    )
    return html_str


def _save_html(
        *, html_str: str, dir_path: str, file_name: str) -> None:
    """
    Save HTML string to file.

    Parameters
    ----------
    html_str : str
        HTML String to save.
    dir_path : str
        Destination directory path to save html.
    file_name : str
        HTML file name. e.g., 'index.html'
    """
    from apysc._file import file_util
    os.makedirs(dir_path, exist_ok=True)
    file_path: str = os.path.join(dir_path, file_name)
    file_util.save_plain_txt(txt=html_str, file_path=file_path)


def _append_expression_to_html_str(*, html_str: str, verbose: int) -> str:
    """
    Append expression strings to a specified HTML string.

    Parameters
    ----------
    html_str : str
        Target HTML string.
    verbose : int, default 1
        The Logging setting.

    Returns
    -------
    html_str : str
        HTML string after appended expressions.
    """
    from apysc._expression import expression_data_util
    from apysc._html import html_const
    from apysc._html import html_util
    _display_info(msg='Reading basic expression data...', verbose=verbose)
    expression: str = expression_data_util.get_current_expression()
    _display_info(msg='Appending common js functions...', verbose=verbose)
    expression = _append_common_js_functions(expression=expression)
    _display_info(
        msg='Appending event handler expressions...', verbose=verbose)
    expression = _append_event_handler_expressions(expression=expression)
    _display_info(msg='Removing unused variables...', verbose=verbose)
    expression = _remove_unused_js_vars(expression=expression)
    _display_info(msg='Removing blank lines...', verbose=verbose)
    expression = _remove_blank_lines(expression=expression)
    expression = html_util.wrap_expression_by_script_tag(
        expression=expression)
    _display_info(msg='Appending indentations...', verbose=verbose)
    expression = html_util.append_indent_to_each_script_line(
        html=expression, indent_num=1)
    entry_point_func_name: str = get_entry_point_func_name()
    expression = expression.replace(
        f'{html_const.SCRIPT_START_TAG}',
        f'{html_const.SCRIPT_START_TAG}'
        f'\nfunction {entry_point_func_name}() {{',
        1)
    expression = expression.replace(
        f'{html_const.SCRIPT_END_TAG}',
        f'}}\n{html_const.SCRIPT_END_TAG}')

    html_str += f'\n{expression}'
    return html_str


def _append_event_handler_expressions(*, expression: str) -> str:
    """
    Append event handler's expressions to a specified string.

    Parameters
    ----------
    expression : str
        Expression string to append event handler's one.

    Returns
    -------
    expression : str
        A result expression string.
    """
    from apysc._expression import expression_data_util
    event_handler_scope_expression: str = \
        expression_data_util.get_current_event_handler_scope_expression()
    expression = f'{expression}\n{event_handler_scope_expression}'
    return expression


def _remove_blank_lines(*, expression: str) -> str:
    """
    Remove blank (break or spaces only) lines from an
    expression string.

    Parameters
    ----------
    expression : str
        Target expression string.

    Returns
    -------
    expression : str
        Expression string that removed blank lines.
    """
    lines: List[str] = expression.splitlines()
    result_lines: List[str] = []
    pattern: Pattern = re.compile(pattern=r'^\s*$')
    for line in lines:
        match: Optional[Match] = pattern.match(string=line)
        if match is not None:
            continue
        result_lines.append(line)
    expression = '\n'.join(result_lines)
    return expression


def _append_common_js_functions(*, expression: str) -> str:
    """
    Append common JavaScript functions (e.g., helper function)
    to a expression string.

    Parameters
    ----------
    expression : str
        Target expression string.

    Returns
    -------
    expression : str
        Expression string which includes fundamental functions.
    """
    from apysc._expression import js_functions
    js_function_strs: List[str] = js_functions.get_js_functions()
    js_function_strs.reverse()
    for js_function in js_function_strs:
        expression = f'{js_function}\n{expression}'
    return expression


def _remove_unused_js_vars(*, expression: str) -> str:
    """
    Remove unused js variables from expression string.

    Parameters
    ----------
    expression : str
        Target js expression string.

    Returns
    -------
    expression : str
        An expression string after removing.
    """
    lines: List[str] = expression.splitlines()
    lines.reverse()
    current_line_idx: int = 0
    while True:
        if current_line_idx >= len(lines):
            break
        var_name: str = _get_var_name_from_line(line=lines[current_line_idx])
        if var_name == '':
            current_line_idx += 1
            continue
        target_js_variable_is_used: bool = _target_js_variable_is_used(
            var_name=var_name, exp_lines=lines)
        if not target_js_variable_is_used:
            del lines[current_line_idx]
            continue
        current_line_idx += 1
    lines.reverse()
    expression = '\n'.join(lines)
    return expression


def _target_js_variable_is_used(
        *, var_name: str, exp_lines: List[str]) -> bool:
    """
    Get a boolean value whether an expression uses a target
    variable in JS expression or not.

    Parameters
    ----------
    var_name : str
        Target variable name.
    exp_lines : list of str
        js expression lines.

    Returns
    -------
    result : bool
        If an expression uses a target variable in JS
        expression, this interface returns True.
    """
    var_pattern: Pattern = re.compile(pattern=rf'var ({var_name}) = ')
    used_pattern_1: Pattern = re.compile(
        pattern=rf'{var_name}[ ;\)\.}},\]\[]')
    used_pattern_2: Pattern = re.compile(
        pattern=rf'{var_name}$')
    for line in exp_lines:
        if '//' in line:
            continue
        if var_name not in line:
            continue
        match: Optional[Match] = var_pattern.search(string=line)
        if match is not None:
            continue
        match = used_pattern_1.search(string=line)
        if match is not None:
            return True
        match = used_pattern_2.search(string=line)
        if match is not None:
            return True
    return False


_VAR_PATTERN: Pattern = re.compile(pattern=r'^var (.+?) = ')


def _get_var_name_from_line(*, line: str) -> str:
    """
    Get a js variable name from a specified line string.

    Parameters
    ----------
    line : str
        Target line string.

    Returns
    -------
    var_name : str
        Result variable name. If line contains `var <any_name> = ...;`
        pattern, then `any_name` will be returned. Or if there is no
        var expression, blank string will be returned.
    """
    line = line.strip()
    match: Optional[Match] = _VAR_PATTERN.search(string=line)
    if match is None:
        return ''
    var_name: str = match.group(1)
    return var_name


def _append_head_to_html_str(
        *, html_str: str, js_lib_dir_path: str,
        embed_js_libs: bool) -> str:
    """
    Append a head tag section to a specified HTML string.

    Parameters
    ----------
    html_str : str
        Target HTML string.
    js_lib_dir_path : str
        JavaScript libraries directory path.
    embed_js_libs: bool, default False
        Option to embed the JavaScript libraries script to the
        output HTML or not.

    Returns
    -------
    html_str : str
        HTML string after appended.
    """
    from apysc._html import html_util
    from apysc._jslib import jslib_util
    html_str = html_util.append_html_to_str(
        to_append_html='<head>', dest_html=html_str, indent_num=0)
    html_str = html_util.append_html_to_str(
        to_append_html='<meta charset="utf-8">',
        dest_html=html_str, indent_num=1)
    jslib_file_names: List[str] = jslib_util.get_jslib_file_names()
    for jslib_file_name in jslib_file_names:
        html_str = _append_jslib_str_to_html(
            html_str=html_str,
            js_lib_dir_path=js_lib_dir_path,
            jslib_file_name=jslib_file_name,
            embed_js_libs=embed_js_libs)
    html_str = html_util.append_html_to_str(
        to_append_html='</head>', dest_html=html_str, indent_num=0)
    return html_str


def _append_jslib_str_to_html(
        *, html_str: str, js_lib_dir_path: str,
        jslib_file_name: str,
        embed_js_libs: bool) -> str:
    """
    Append JavaScript libraries script string to HTML.

    Parameters
    ----------
    html_str : str
        A HTML string to be appended.
    js_lib_dir_path : str
        Target libraries directory path.
        If the `embed_js_libs` option is True, then this setting
        will be ignored.
    jslib_file_name : str
        Target JavaScript library file name.
    embed_js_libs: bool, default False
        Option to embed the JavaScript libraries script to the
        output HTML or not.

    Returns
    -------
    html_str : str
        Result HTML string.
    """
    from apysc._html import html_util
    from apysc._jslib import jslib_util
    if not embed_js_libs:
        html_str = html_util.append_html_to_str(
            to_append_html=(
                '<script type="text/javascript" '
                f'src="{js_lib_dir_path}{jslib_file_name}"></script>'),
            dest_html=html_str, indent_num=1)
        return html_str

    js_script: str = (
        '<script type="text/javascript">'
        f'\n  {jslib_util.read_jslib_str(jslib_name=jslib_file_name)}'
        '\n  </script>'
    )
    html_str = html_util.append_html_to_str(
        to_append_html=js_script, dest_html=html_str, indent_num=1)
    return html_str


def _export_js_libs(
        *, dest_dir_path: str, skip_js_lib_exporting: bool) -> List[str]:
    """
    Export JavaScript libraries to a specified directory.

    Parameters
    ----------
    dest_dir_path : str
        Directory path to export JavaScript libraries.
    skip_js_lib_exporting : bool
        If True is set, then JavaScript libraries will not be exported.

    Returns
    -------
    saved_js_file_paths : str
        Saved JavaScript file paths.
    """
    if skip_js_lib_exporting:
        return []

    from apysc._jslib import jslib_util
    jslib_file_names: List[str] = jslib_util.get_jslib_file_names()
    saved_js_file_paths: List[str] = []
    for jslib_file_name in jslib_file_names:
        saved_js_file_path: str = jslib_util.export_jslib_to_specified_dir(
            dest_dir_path=dest_dir_path, jslib_name=jslib_file_name)
        saved_js_file_paths.append(saved_js_file_path)
    return saved_js_file_paths



"""This module is for the translation mapping data of the
following document:

Document file: dictionary.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Dictionary':
    '',

    'This page explains the `Dictionary` class.\n\nBefore reading on, maybe it is helpful to read the following page:\n\n- [Why the apysc library doesn\'t use the Python built-in data type](why_apysc_doesnt_use_python_builtin_data_type.md)':  # noqa
    '',

    '## What is the Dictionary?':
    '',

    'The `Dictionary` class is the apysc dictionary class. It behaves like the Python built-in `dict` value.':  # noqa
    '',

    '## Constructor method':
    '',

    'The `Dictionary` class constructor method requires a Python built-in `dict` or `Dictionary` value:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\ndict_1: ap.Dictionary = ap.Dictionary({\'a\': 10})\nassert dict_1 == {\'a\': 10}\n\ndict_2: ap.Dictionary = ap.Dictionary(dict_1)\nassert dict_1 == dict_2\n```':  # noqa
    '',

    '':
    '',

    '## Value setter interface':
    '',

    'A `Dictionary` value can be updated by indexing, like the Python built-in `dict` value:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\ndict_1: ap.Dictionary = ap.Dictionary({\'a\': 10})\ndict_1[\'a\'] = 20\nassert dict_1 == {\'a\': 20}\n```':  # noqa
    '',

    '':
    '',

    '## Value getter interface':
    '',

    'A `Dictionary` value also can be retrieved by indexing:':
    '',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\ndict_1: ap.Dictionary = ap.Dictionary({\'a\': int_1})\nint_2: ap.Int = dict_1[\'a\']\nassert isinstance(int_2, ap.Int)\nassert int_2 == 10\n```':  # noqa
    '',

    '':
    '',

    '## Notes of the getter interface':
    '',

    'If a `Dictionary` value doesn\'t have the specified key, a retrieved value type becomes the `AnyValue` type. This behavior occasionally is helpful when a `Dictionary` value is updated dynamically (e.g., updating by the JavaScript event handler).':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\ndict_1: ap.Dictionary = ap.Dictionary({\'a\': int_1})\nretrieved_val: ap.AnyValue = dict_1[\'b\']\nassert isinstance(retrieved_val, ap.AnyValue)\n```':  # noqa
    '',

    '':
    '',

    '## Value deletion interface':
    '',

    'A `Dictionary` value can be deleted by the `del` statement, as follows:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\ndict_1: ap.Dictionary = ap.Dictionary({\'a\': int_1})\ndel dict_1[\'a\']\nassert dict_1 == {}\n```':  # noqa
    '',

    '':
    '',

    '## Dictionary class constructor API':
    '',

    '<!-- Docstring: apysc._type.dictionary.Dictionary.__init__ -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `__init__(self, value:Union[Dict[~_K, ~_V], _ForwardRef(\'Dictionary\')]) -> None`<hr>\n\n**[Interface summary]** Dictionary class for the apysc library.<hr>\n\n**[Parameters]**\n\n- `value`: dict or Dictionary\n  - Initial dictionary value.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> dictionary: ap.Dictionary = ap.Dictionary({\'a\': 10})\n>>> dictionary\nDictionary({\'a\': 10})\n\n>>> dictionary[\'a\']\n10\n\n>>> dictionary[\'b\'] = 20\n>>> dictionary[\'b\']\n20\n```':  # noqa
    '',

    '<hr>\n\n**[References]**\n\n- [Dictionary class generic type settings document](https://simon-ritchie.github.io/apysc/dictionary_generic.html)':  # noqa
    '',

    '## value attribute API':
    '',

    '<!-- Docstring: apysc._type.dictionary.Dictionary.value -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface summary]** Get a current dict value.<hr>\n\n**[Returns]**\n\n- `value`: dict\n  - Current dict value.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> dictionary: ap.Dictionary = ap.Dictionary({})\n>>> dictionary.value = {\'a\': 10}\n>>> dictionary.value\n{\'a\': 10}\n```':  # noqa
    '',

    '<hr>\n\n**[References]**\n\n- [apysc fundamental data classes value interface](https://simon-ritchie.github.io/apysc/fundamental_data_classes_value_interface.html)':  # noqa
    '',

}



"""This module is for the translation conversion
of the documents.
"""

from typing import Dict
from typing import List
from typing import Union

from apysc._lint_and_doc.docs_lang import Lang


class _TranslationMappingNotFound(Exception):
    pass


_EACH_LANG_HEADING_INFO_FORMAT: Dict[Lang, str] = {
    Lang.JP:
    '<span class="inconspicuous-txt">※この翻訳ドキュメントはスクリプト'
    'によって出力・同期されています。内容が怪しそうな場合はGitHubに'
    'issueを追加したり[英語の原文]({source_doc_path})を'
    '確認してください。</span>',
}


def apply_translation_to_doc(
        *, md_file_path: str, lang: Lang) -> str:
    """
    Apply a translation mapping to a specified markdown
    document. This interface saves a translated file
    with a file name with a language suffix (e.g.,
    jp_your_doc.md).

    Parameters
    ----------
    md_file_path : str
        A source markdown file path.
    lang : Lang
        A target language setting.

    Returns
    -------
    translated_file_path : str
        A translated document file path.
    """
    from apysc._file import file_util
    from apysc._lint_and_doc import translation_mapping_utils
    from apysc._lint_and_doc.document_text_split_util import BodyText
    from apysc._lint_and_doc.document_text_split_util import CodeBlock
    from apysc._lint_and_doc.document_text_split_util import Heading
    from apysc._lint_and_doc.document_text_split_util import \
        split_markdown_document
    markdown: str = file_util.read_txt(file_path=md_file_path)
    splitted_values: List[Union[Heading, BodyText, CodeBlock]] = \
        split_markdown_document(markdown_txt=markdown)
    mapping_data: Dict[str, str] = translation_mapping_utils.\
        read_mapping_data(src_doc_file_path=md_file_path, lang=lang)
    keys: List[str] = translation_mapping_utils.\
        convert_splitted_values_to_keys(splitted_values=splitted_values)
    translated_doc: str = ''
    translated_doc = _add_heading_info_if_exists(
        translated_doc=translated_doc, lang=lang,
        md_file_path=md_file_path)
    for key in keys:
        key_: str = translation_mapping_utils.\
            remove_escaping_from_key_or_value(key_or_val=key)

        if translation_mapping_utils.is_translation_skipping_key(key=key_):
            continue

        if translated_doc != '':
            translated_doc += '\n\n'

        is_mapping_unnecessary_key: bool = translation_mapping_utils.\
            is_mapping_unnecessary_key(key=key_)
        if is_mapping_unnecessary_key:
            translated_doc += key_
            continue

        translated_str: str = mapping_data.get(key_, '')
        translated_doc += translated_str
        _validate_translated_str_is_not_blank(
            translated_str=translated_str, key=key,
            md_file_path=md_file_path)

    translated_file_path: str = translation_mapping_utils.\
        get_translated_file_path_from_src_path(
            source_doc_path=md_file_path, lang=lang)
    file_util.save_plain_txt(
        txt=translated_doc, file_path=translated_file_path)

    return translated_file_path


def _add_heading_info_if_exists(
        *, translated_doc: str, lang: Lang,
        md_file_path: str) -> str:
    """
    Add a target language's heading information text
    to a specified translated text.

    Parameters
    ----------
    translated_doc : str
        A target translated document's text.
    lang : Lang
        A target language.
    md_file_path : str
        A target source document path.

    Returns
    -------
    translated_doc : str
        A result translated document's text.
        This interface directly returns argument value
        if there is no heading information setting.
    """
    heading_info_format: str = _EACH_LANG_HEADING_INFO_FORMAT.get(
        lang, '')
    if heading_info_format == '':
        return translated_doc
    heading_info_format = heading_info_format.format(
        source_doc_path=md_file_path,
    )
    translated_doc += heading_info_format
    return translated_doc


def _validate_translated_str_is_not_blank(
        *, translated_str: str, key: str,
        md_file_path: str) -> None:
    """
    Validate whether a translated string is not a blank string.

    Parameters
    ----------
    translated_str : str
        A target translated string.
    key : str
        A target key string (original source text).
    md_file_path : str
        A source markdown file path.

    Raises
    ------
    _TranslationMappingNotFound
        If a specified translated string is a blank string.
    """
    if translated_str != '':
        return
    raise _TranslationMappingNotFound(
        'There is no translation mapping.'
        f'\nOriginal source text: {key}'
        f'\nSource document path: {md_file_path}')



"""This module is for the translation mapping data of the
following document:

Document file: timer_reset.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Timer class reset interface':
    '',

    'This page explains the `Timer` class `reset` method interface.':
    '',

    '## What interface is this?':
    '',

    'The `reset` method interface resets a timer count and stops it.':
    '',

    '## Basic usage':
    '',

    'The `reset` method has no arguments.\n\nThe following example rotates the rectangle 90 degrees (`repeat_count=90`) by the first-timer and then reset and start the second one. The second-timer resets and restart the first-timer after 1 second.':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\nap.Stage(\n    stage_width=150, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\n\nclass _RectOptions(TypedDict):\n    rectangle: ap.Rectangle\n\n\nclass _TimerOptions(TypedDict):\n    timer: ap.Timer\n\n\ndef on_first_timer(e: ap.TimerEvent, options: _RectOptions) -> None:\n    """\n    The handler that the first-timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = options[\'rectangle\']\n    rectangle.rotation_around_center += 1\n\n\ndef on_first_timer_complete(\n        e: ap.TimerEvent, options: _TimerOptions) -> None:\n    """\n    The handler that the first-timer calls when completed.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    timer_2: ap.Timer = options[\'timer\']\n    timer_2.reset()\n    timer_2.start()\n\n\ndef on_second_timer(e: ap.TimerEvent, options: _TimerOptions) -> None:\n    """\n    The handler that the second timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    timer_1: ap.Timer = options[\'timer\']\n    timer_1.reset()\n    timer_1.start()\n\n\noptions_1: _RectOptions = {\'rectangle\': rectangle}\ntimer_1: ap.Timer = ap.Timer(\n    handler=on_first_timer, delay=ap.FPS.FPS_60, repeat_count=90,\n    options=options_1)\n\noptions_2: _TimerOptions = {\'timer\': timer_1}\ntimer_2: ap.Timer = ap.Timer(\n    handler=on_second_timer, delay=1000, repeat_count=1,\n    options=options_2)\noptions_2 = {\'timer\': timer_2}\ntimer_1.timer_complete(\n    on_first_timer_complete, options=options_2)\ntimer_1.start()\n\nap.save_overall_html(\n    dest_dir_path=\'timer_reset_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/timer_reset_basic_usage/index.html" width="150" height="150"></iframe>':  # noqa
    '',

    '## reset API':
    '',

    '<!-- Docstring: apysc._time.timer.Timer.reset -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `reset(self) -> None`<hr>\n\n**[Interface summary]** Reset the timer count and stop this timer.<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> from typing_extensions import TypedDict\n>>> import apysc as ap\n>>> class RectOptions(TypedDict):\n...     rectangle: ap.Rectangle\n>>> def on_timer(e: ap.TimerEvent, options: RectOptions) -> None:\n...     rectangle: ap.Rectangle = options[\'rectangle\']\n...     rectangle.x += 1\n...     with ap.If(rectangle.x > 100):\n...         timer: ap.Timer = e.this\n...         timer.reset()\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> options: RectOptions = {\'rectangle\': rectangle}\n>>> _ = ap.Timer(\n...     on_timer, delay=33.3, options=options).start()\n```':  # noqa
    '',

}



"""Implementations for Graphics class.
"""

from typing import List
from typing import Optional
from typing import Union

from apysc._display import circle as _circle
from apysc._display import ellipse as _ellipse
from apysc._display import line as _line
from apysc._display import path as _path
from apysc._display import polygon as _polyg
from apysc._display import polyline as _polyline
from apysc._display import sprite
from apysc._display.begin_fill_interface import BeginFillInterface
from apysc._display.child_interface import ChildInterface
from apysc._display.display_object import DisplayObject
from apysc._display.graphics_clear_interface import GraphicsClearInterface
from apysc._display.line_style_interface import LineStyleInterface
from apysc._display.rectangle import Rectangle
from apysc._geom.path_data_base import PathDataBase
from apysc._geom.point2d import Point2D
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.array import Array
from apysc._type.int import Int
from apysc._type.variable_name_interface import VariableNameInterface


class Graphics(
        DisplayObject,
        BeginFillInterface, LineStyleInterface, VariableNameInterface,
        GraphicsClearInterface, ChildInterface):
    """
    Create an object that has each vector graphics interface.

    References
    ----------
    - Graphics document
        - https://simon-ritchie.github.io/apysc/graphics.html

    Examples
    --------
    >>> import apysc as ap
    >>> stage: ap.Stage = ap.Stage()
    >>> sprite: ap.Sprite = ap.Sprite()
    >>> sprite.graphics.begin_fill(color='#0af')
    >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
    ...     x=50, y=50, width=50, height=50)
    >>> rectangle.x
    Int(50)

    >>> circle: ap.Circle = sprite.graphics.draw_circle(
    ...     x=100, y=100, radius=50)
    >>> circle.x
    Int(100)
    """

    _current_line: Optional['_polyline.Polyline'] = None

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Graphics')
    def __init__(
            self, *, parent: 'sprite.Sprite',
            variable_name: Optional[str] = None) -> None:
        """
        Create an object that has each vector graphics interface.

        Parameters
        ----------
        parent : Sprite
            A parent instance.
        variable_name : str or None, default None
            Variable name to set. Specified only when
            a subclass instantiation.

        References
        ----------
        - Graphics document
            - https://simon-ritchie.github.io/apysc/graphics.html
        """
        import apysc as ap
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        from apysc._validation import display_validation

        display_validation.validate_sprite(sprite=parent)
        self.parent_sprite: ap.Sprite = parent
        if variable_name is None:
            variable_name = expression_variables_util.\
                get_next_variable_name(type_name=var_names.GRAPHICS)
        super(Graphics, self).__init__(variable_name=variable_name)
        self._fill_color = ap.String('')
        self._fill_alpha = ap.Number(1.0)
        self._line_color = ap.String('')
        self._line_alpha = ap.Number(1.0)
        self._line_thickness = ap.Int(1.0)
        self._children = ap.Array([])
        self._append_constructor_expression()
        self.parent_sprite.add_child(self)
        self._set_overflow_visible_setting()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Graphics')
    def _append_constructor_expression(self) -> None:
        """
        Append constructor expression.
        """
        import apysc as ap
        stage_name: str = self.parent_sprite.stage.variable_name
        expression: str = (
            f'var {self.variable_name} = {stage_name}.nested();'
        )
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Graphics')
    def draw_rect(
            self, x: Union[int, Int],
            y: Union[int, Int],
            width: Union[int, Int],
            height: Union[int, Int]) -> Rectangle:
        """
        Draw a rectangle vector graphics.

        Parameters
        ----------
        x : Int or int
            X position to start drawing.
        y : Int or int
            Y position to start drawing.
        width : Int or int
            Rectangle width.
        height : Int or int
            Rectangle height.

        Returns
        -------
        rectangle : Rectangle
            Created rectangle.

        References
        ----------
        - Graphics draw_rect interface document
            - https://simon-ritchie.github.io/apysc/graphics_draw_rect.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> rectangle.x
        Int(50)

        >>> rectangle.width
        Int(50)

        >>> rectangle.fill_color
        String('#00aaff')
        """
        rectangle: Rectangle = Rectangle(
            parent=self, x=x, y=y, width=width, height=height)
        self.add_child(child=rectangle)
        return rectangle

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Graphics')
    def draw_round_rect(
            self, x: Union[int, Int],
            y: Union[int, Int],
            width: Union[int, Int],
            height: Union[int, Int],
            ellipse_width: Union[int, Int],
            ellipse_height: Union[int, Int]) -> Rectangle:
        """
        Draw a rounded rectangle vector graphics.

        Parameters
        ----------
        x : Int or int
            X-coordinate to start drawing.
        y : Int or int
            Y-coordinate to start drawing.
        width : Int or int
            Rectangle width.
        height : Int or int
            Rectangle height.
        ellipse_width : Int or int
            Ellipse width of the rectangle corner.
        ellipse_height : Int or int
            Ellipse height of the rectangle corner.

        Returns
        -------
        rectangle : Rectangle
            Created rectangle.

        References
        ----------
        - Graphics draw_round_rect interface document
            - https://simon-ritchie.github.io/apysc/graphics_draw_round_rect.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> round_rect: ap.Rectangle = sprite.graphics.draw_round_rect(
        ...     x=50, y=50, width=50, height=50,
        ...     ellipse_width=10, ellipse_height=15)
        >>> round_rect.ellipse_width
        Int(10)

        >>> round_rect.ellipse_height
        Int(15)
        """
        import apysc as ap
        rectangle: Rectangle = Rectangle(
            parent=self, x=x, y=y, width=width, height=height)
        if isinstance(ellipse_width, int):
            ellipse_width = ap.Int(ellipse_width)
        if isinstance(ellipse_height, int):
            ellipse_height = ap.Int(ellipse_height)
        rectangle.ellipse_width = ellipse_width
        rectangle.ellipse_height = ellipse_height
        self.add_child(child=rectangle)
        return rectangle

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Graphics')
    def draw_circle(
            self,
            x: Union[int, Int],
            y: Union[int, Int],
            radius: Union[int, Int]) -> '_circle.Circle':
        """
        Draw a circle vector graphics.

        Parameters
        ----------
        x : Int or int
            X-coordinate of the circle center.
        y : Int or int
            Y-coordinate of the circle center.
        radius : Int or int
            Circle radius.

        Returns
        -------
        circle : Circle
            Created circle graphics instance.

        References
        ----------
        - Graphics draw_circle interface document
            - https://simon-ritchie.github.io/apysc/graphics_draw_circle.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> circle: ap.Circle = sprite.graphics.draw_circle(
        ...     x=100, y=100, radius=50)
        >>> circle.x
        Int(100)

        >>> circle.y
        Int(100)

        >>> circle.radius
        Int(50)

        >>> circle.fill_color
        String('#00aaff')
        """
        circle: _circle.Circle = _circle.Circle(
            parent=self, x=x, y=y, radius=radius)
        self.add_child(child=circle)
        return circle

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Graphics')
    def draw_ellipse(
            self,
            x: Union[int, Int],
            y: Union[int, Int],
            width: Union[int, Int],
            height: Union[int, Int]) -> '_ellipse.Ellipse':
        """
        Draw an ellipse vector graphic.

        Parameters
        ----------
        x : Int or int
            X-coordinate of the ellipse center.
        y : Int or int
            Y-coordinate of the ellipse center.
        width : Int or int
            Ellipse width.
        height : Int or int
            Ellipse height.

        Returns
        -------
        ellipse : Ellipse
            Created ellipse graphics instance.

        References
        ----------
        - Graphics draw_ellipse interface
            - https://simon-ritchie.github.io/apysc/graphics_draw_ellipse.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> ellipse: ap.Ellipse = sprite.graphics.draw_ellipse(
        ...     x=100, y=100, width=100, height=50)
        >>> ellipse.x
        Int(100)

        >>> ellipse.y
        Int(100)

        >>> ellipse.width
        Int(100)

        >>> ellipse.height
        Int(50)

        >>> ellipse.fill_color
        String('#00aaff')
        """
        ellipse: _ellipse.Ellipse = _ellipse.Ellipse(
            parent=self, x=x, y=y, width=width, height=height)
        self.add_child(child=ellipse)
        return ellipse

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Graphics')
    def line_to(
            self, x: Union[int, Int],
            y: Union[int, Int]) -> '_polyline.Polyline':
        """
        Draw a line from previous point to specified point (initial
        point is x = 0, y = 0).

        Parameters
        ----------
        x : Int or int
            X destination point to draw a line.
        y : Int or int
            Y destination point to draw a line.

        Returns
        -------
        line : Polyline
            Line graphics instance.

        References
        ----------
        - Graphics move_to and line_to interfaces document
            - https://simon-ritchie.github.io/apysc/graphics_move_to_and_line_to.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=5)
        >>> line_1: ap.Polyline = sprite.graphics.move_to(x=50, y=50)
        >>> line_2: ap.Polyline = sprite.graphics.line_to(x=150, y=50)
        >>> line_3: ap.Polyline = sprite.graphics.line_to(x=50, y=150)
        >>> line_1 == line_2 == line_3
        True

        >>> line_1.line_color
        String('#ffffff')

        >>> line_1.line_thickness
        Int(5)
        """
        import apysc as ap
        if self._current_line is None:
            self._current_line = _polyline.Polyline(
                parent=self,
                points=ap.Array([Point2D(x=0, y=0), Point2D(x=x, y=y)]))
            self.add_child(self._current_line)
        else:
            self._current_line.append_line_point(x=x, y=y)
        return self._current_line

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Graphics')
    def move_to(
            self,
            x: Union[int, Int],
            y: Union[int, Int]) -> '_polyline.Polyline':
        """
        Move a line position to a specified point.

        Parameters
        ----------
        x : Int or int
            X destination point to move.
        y : Int or int
            Y destination point to move.

        Returns
        -------
        line : Polyline
            Line graphics instance.

        References
        ----------
        - Graphics move_to and line_to interfaces document
            - https://simon-ritchie.github.io/apysc/graphics_move_to_and_line_to.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=5)
        >>> line_1: ap.Polyline = sprite.graphics.move_to(x=50, y=50)
        >>> line_2: ap.Polyline = sprite.graphics.line_to(x=150, y=50)
        >>> line_1 == line_2
        True

        >>> line_1.line_color
        String('#ffffff')

        >>> line_1.line_thickness
        Int(5)
        """
        import apysc as ap
        polyline: _polyline.Polyline = _polyline.Polyline(
            parent=self, points=ap.Array([Point2D(x=x, y=y)]))
        self._current_line = polyline
        return polyline

    def _reset_each_line_settings(self) -> None:
        """
        Reset each line settings (e.g., LineDotSetting, LineDashSetting,
        and so on).

        Notes
        -----
        This interface does not append an expression.
        """
        self._line_dot_setting = None
        self._line_dash_setting = None
        self._line_round_dot_setting = None
        self._line_dash_dot_setting = None

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Graphics')
    def draw_line(
            self,
            x_start: Union[int, Int],
            y_start: Union[int, Int],
            x_end: Union[int, Int],
            y_end: Union[int, Int]) -> '_line.Line':
        """
        Draw a normal line vector graphic.

        Notes
        -----
        - This interface ignores line settings, like
            the `LineDotSetting`, `LineDashSetting`.

        Parameters
        ----------
        x_start : Int or int
            Line start x-coordinate.
        y_start : Int or int
            Line start y-coordinate.
        x_end : Int or int
            Line end x-coordinate.
        y_end : Int or int
            Line end y-coordinate.

        Returns
        -------
        line : Line
            Created line graphics instance.

        References
        ----------
        - Graphics draw_line interface document
            - https://simon-ritchie.github.io/apysc/graphics_draw_line.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=5)
        >>> line: ap.Line = sprite.graphics.draw_line(
        ...     x_start=50, y_start=50, x_end=150, y_end=50)
        >>> line.line_color
        String('#ffffff')

        >>> line.line_thickness
        Int(5)
        """
        snapshot_name: str = self._get_next_snapshot_name()
        self._run_all_make_snapshot_methods(snapshot_name=snapshot_name)
        self._reset_each_line_settings()
        line: _line.Line = _line.Line(
            parent=self,
            start_point=Point2D(x=x_start, y=y_start),
            end_point=Point2D(x=x_end, y=y_end))
        self._run_all_revert_methods(snapshot_name=snapshot_name)
        self.add_child(child=line)
        return line

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Graphics')
    def draw_dotted_line(
            self,
            x_start: Union[int, Int],
            y_start: Union[int, Int],
            x_end: Union[int, Int],
            y_end: Union[int, Int],
            dot_size: Union[int, Int]) -> '_line.Line':
        """
        Draw a dotted line vector graphics.

        Notes
        -----
        - This interface ignores line settings, like the
            `LineDashSetting`, except `LineDotSetting`.

        Parameters
        ----------
        x_start : Int or int
            Line start x-coordinate.
        y_start : Int or int
            Line start y-coordinate.
        x_end : Int or int
            Line end x-coordinate.
        y_end : Int or int
            Line end y-coordinate.
        dot_size : Int or int
            Dot size.

        Returns
        -------
        line : Line
            Created line graphics instance.

        References
        ----------
        - Graphics draw_dotted_line interface document
            - https://simon-ritchie.github.io/apysc/graphics_draw_dotted_line.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=5)
        >>> line: ap.Line = sprite.graphics.draw_dotted_line(
        ...     x_start=50, y_start=50, x_end=150, y_end=50, dot_size=5)
        >>> line.line_color
        String('#ffffff')

        >>> line.line_thickness
        Int(5)

        >>> line.line_dot_setting.dot_size
        Int(5)
        """
        import apysc as ap
        snapshot_name: str = self._get_next_snapshot_name()
        self._run_all_make_snapshot_methods(snapshot_name=snapshot_name)
        self._reset_each_line_settings()
        self._line_dot_setting = ap.LineDotSetting(dot_size=dot_size)
        line: _line.Line = _line.Line(
            parent=self,
            start_point=Point2D(x=x_start, y=y_start),
            end_point=Point2D(x=x_end, y=y_end))
        self._run_all_revert_methods(snapshot_name=snapshot_name)
        self.add_child(child=line)
        return line

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Graphics')
    def draw_dashed_line(
            self,
            x_start: Union[int, Int],
            y_start: Union[int, Int],
            x_end: Union[int, Int],
            y_end: Union[int, Int],
            dash_size: Union[int, Int],
            space_size: Union[int, Int]) -> '_line.Line':
        """
        Draw a dashed line vector graphics.

        Notes
        -----
        - This interface ignores line settings, like
        the `LineDotSetting`, except `LineDashSetting`.

        Parameters
        ----------
        x_start : Int or int
            Line start x-coordinate.
        y_start : Int or int
            Line start y-coordinate.
        x_end : Int or int
            Line end x-coordinate.
        y_end : Int or int
            Line end y-coordinate.
        dash_size : Int or int
            Dash size.
        space_size : Int or int
            Blank space size between dashes.

        Returns
        -------
        line : Line
            Created line graphics instance.

        References
        ----------
        - Graphics draw_dashed_line interface document
            - https://simon-ritchie.github.io/apysc/graphics_draw_dashed_line.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=5)
        >>> line: ap.Line = sprite.graphics.draw_dashed_line(
        ...     x_start=50, y_start=50, x_end=150, y_end=50,
        ...     dash_size=5, space_size=2)
        >>> line.line_color
        String('#ffffff')

        >>> line.line_dash_setting.dash_size
        Int(5)

        >>> line.line_dash_setting.space_size
        Int(2)
        """
        import apysc as ap
        snapshot_name: str = self._get_next_snapshot_name()
        self._run_all_make_snapshot_methods(snapshot_name=snapshot_name)
        self._reset_each_line_settings()
        self._line_dash_setting = ap.LineDashSetting(
            dash_size=dash_size, space_size=space_size)
        line: _line.Line = _line.Line(
            parent=self,
            start_point=Point2D(x=x_start, y=y_start),
            end_point=Point2D(x=x_end, y=y_end))
        self._run_all_revert_methods(snapshot_name=snapshot_name)
        self.add_child(child=line)
        return line

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Graphics')
    def draw_round_dotted_line(
            self,
            x_start: Union[int, Int],
            y_start: Union[int, Int],
            x_end: Union[int, Int],
            y_end: Union[int, Int],
            round_size: Union[int, Int],
            space_size: Union[int, Int]) -> '_line.Line':
        """
        Draw a round-dotted line vector graphics.

        Notes
        -----
        This interface ignores line settings, like the
        `LineDotSetting`, except `LineRoundDotSetting`.

        Parameters
        ----------
        x_start : Int or int
            Line start x-coordinate.
        y_start : Int or int
            Line start y-coordinate.
        x_end : Int or int
            Line end x-coordinate.
        y_end : Int or int
            Line end y-coordinate.
        round_size : Int or int
            Dot round size.
        space_size : Int or int
            Blank space size between dots.

        Returns
        -------
        line : Line
            Created line graphics instance.

        References
        ----------
        - Graphics draw_round_dotted_line interface document
            - https://simon-ritchie.github.io/apysc/graphics_draw_round_dotted_line.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=5)
        >>> line: ap.Line = sprite.graphics.draw_round_dotted_line(
        ...    x_start=50, y_start=50, x_end=150, y_end=50,
        ...    round_size=6, space_size=3)
        >>> line.line_color
        String('#ffffff')

        >>> line.line_round_dot_setting.round_size
        Int(6)

        >>> line.line_round_dot_setting.space_size
        Int(3)
        """
        import apysc as ap
        snapshot_name: str = self._get_next_snapshot_name()
        self._run_all_make_snapshot_methods(snapshot_name=snapshot_name)
        self._reset_each_line_settings()
        self._line_round_dot_setting = ap.LineRoundDotSetting(
            round_size=round_size, space_size=space_size)
        line: _line.Line = _line.Line(
            parent=self,
            start_point=Point2D(x=x_start, y=y_start),
            end_point=Point2D(x=x_end, y=y_end))
        self._run_all_revert_methods(snapshot_name=snapshot_name)
        self.add_child(child=line)
        return line

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Graphics')
    def draw_dash_dotted_line(
            self,
            x_start: Union[int, Int],
            y_start: Union[int, Int],
            x_end: Union[int, Int],
            y_end: Union[int, Int],
            dot_size: Union[int, Int],
            dash_size: Union[int, Int],
            space_size: Union[int, Int]) -> '_line.Line':
        """
        Draw a dash-dotted (1-dot chain) line vector graphics.

        Parameters
        ----------
        x_start : Int or int
            Line start x-coordinate.
        y_start : Int or int
            Line start y-coordinate.
        x_end : Int or int
            Line end x-coordinate.
        y_end : Int or int
            Line end y-coordinate.
        dot_size : Int or int
            Dot size.
        dash_size : Int or int
            Dash size.
        space_size : Int or int
            Blank space size between dots and dashes.

        Returns
        -------
        line : Line
            Created line graphics instance.

        References
        ----------
        - Graphics draw_dash_dotted_line interface document
            - https://simon-ritchie.github.io/apysc/graphics_draw_dash_dotted_line.html # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=5)
        >>> line: ap.Line = sprite.graphics.draw_dash_dotted_line(
        ...    x_start=50, y_start=50, x_end=150, y_end=50,
        ...    dot_size=2, dash_size=5, space_size=3)
        >>> line.line_color
        String('#ffffff')

        >>> line.line_dash_dot_setting.dot_size
        Int(2)

        >>> line.line_dash_dot_setting.dash_size
        Int(5)

        >>> line.line_dash_dot_setting.space_size
        Int(3)
        """
        import apysc as ap
        snapshot_name: str = self._get_next_snapshot_name()
        self._run_all_make_snapshot_methods(snapshot_name=snapshot_name)
        self._reset_each_line_settings()
        self._line_dash_dot_setting = ap.LineDashDotSetting(
            dot_size=dot_size,
            dash_size=dash_size,
            space_size=space_size)
        line: _line.Line = _line.Line(
            parent=self,
            start_point=Point2D(x=x_start, y=y_start),
            end_point=Point2D(x=x_end, y=y_end))
        self._run_all_revert_methods(snapshot_name=snapshot_name)
        self.add_child(child=line)
        return line

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Graphics')
    def draw_polygon(
            self,
            points: Union[List[Point2D], Array[Point2D]]) -> '_polyg.Polygon':
        """
        Draw a polygon vector graphic. This interface is similar
        to the Polyline class (created by `move_to` or `line_to`).
        But unlike that, this interface connects the last point
        and the start point.

        Parameters
        ----------
        points : list of Point2D or Array.
            Polygon vertex points.

        Returns
        -------
        polygon : Polygon
            Created polygon graphics instance.

        References
        ----------
        - Graphics draw_polygon interface document
            - https://simon-ritchie.github.io/apysc/graphics_draw_polygon.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> polygon: ap.Polygon = sprite.graphics.draw_polygon(
        ...     points=[
        ...         ap.Point2D(x=25, y=0),
        ...         ap.Point2D(x=0, y=50),
        ...         ap.Point2D(x=50, y=50),
        ...     ])
        >>> polygon.fill_color
        String('#00aaff')
        """
        import apysc as ap
        if isinstance(points, list):
            points = ap.Array(points)
        polygon: _polyg.Polygon = _polyg.Polygon(
            parent=self, points=points)
        self.add_child(polygon)
        return polygon

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Graphics')
    def draw_path(
            self, path_data_list: List[PathDataBase]) -> '_path.Path':
        """
        Draw a path vector graphics.

        Parameters
        ----------
        path_data_list : list of PathDataBase
            Target path data settings, such as the ap.PathData.MoveTo.

        Returns
        -------
        path : Path
            Created path graphics instance.

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=3)
        >>> path: ap.Path = sprite.graphics.draw_path(
        ...     path_data_list=[
        ...         ap.PathMoveTo(x=0, y=50),
        ...         ap.PathBezier2D(
        ...             control_x=50, control_y=0,
        ...             dest_x=100, dest_y=50),
        ...     ])
        """
        path: _path.Path = _path.Path(
            parent=self, path_data_list=path_data_list)
        self.add_child(path)
        return path

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of
        debugging).

        Returns
        -------
        repr_str : str
            Type name and variable name will be set
            (e.g., `Graphics('<variable_name>')`).
        """
        repr_str: str = f"Graphics('{self.variable_name}')"
        return repr_str



"""This module is for the translation mapping data of the
following document:

Document file: animation_time.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# animation_time interface':
    '',

    'This page explains the `animation_time` method interface.':
    '',

    '## What interface is this?':
    '',

    'The `animation_time` interface returns the current animation elapsed time in milliseconds (`Number` type value).':  # noqa
    '',

    '## Basic usage':
    '',

    'The following example sets the x-coordinate animation of the rectangle and the 1-second interval timer to display an animation\'s current elapsed time to console (please press the F12 key).':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass _RectOptions(TypedDict):\n    rectangle: ap.Rectangle\n\n\ndef on_timer(e: ap.TimerEvent, options: _RectOptions) -> None:\n    """\n    The handler that the animation calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = options[\'rectangle\']\n    ap.trace(\'Animation elapsed time:\', rectangle.animation_time())\n\n\nap.Stage(\n    stage_width=500, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\n\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nrectangle.animation_x(x=400, duration=10000).start()\n\noptions: _RectOptions = {\'rectangle\': rectangle}\nap.Timer(on_timer, delay=1000, options=options).start()\n\nap.save_overall_html(\n    dest_dir_path=\'animation_time_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/animation_time_basic_usage/index.html" width="500" height="150"></iframe>':  # noqa
    '',

    '## animation_time property API':
    '',

    '<!-- Docstring: apysc._animation.animation_time_interface.AnimationTimeInterface.animation_time -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `animation_time(self) -> apysc._type.number.Number`<hr>\n\n**[Interface summary]** Get an animation elapsed millisecond.<hr>\n\n**[Returns]**\n\n- `elapsed_time`: Number\n  - An animation elapsed millisecond.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> from typing_extensions import TypedDict\n>>> import apysc as ap\n>>> class RectOptions(TypedDict):\n...     rectangle: ap.Rectangle\n>>> def on_timer(\n...         e: ap.TimerEvent,\n...         options: RectOptions) -> None:\n...     rectangle: ap.Rectangle = options[\'rectangle\']\n...     animation_time: ap.Number = rectangle.animation_time()\n...     ap.trace(\'animation_time:\', animation_time)\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> _ = rectangle.animation_x(\n...     x=100,\n...     duration=1500,\n...     easing=ap.Easing.EASE_OUT_QUINT,\n... ).start()\n>>> options: RectOptions = {\'rectangle\': rectangle}\n>>> ap.Timer(\n...     on_timer, delay=ap.FPS.FPS_60,\n...     options=options).start()\n```':  # noqa
    '',

}



"""Implementation of VariableNameInterface class.
"""

from typing import List


class VariableNameInterface:

    _variable_name: str
    _variable_name_history: List[str]

    @property
    def variable_name(self) -> str:
        """
        Get a js variable name of this instance.

        Returns
        -------
        variable_name : str
            A js variable name of this instance.
        """
        if not hasattr(self, '_variable_name'):
            return ''
        return self._variable_name

    @variable_name.setter
    def variable_name(self, variable_name: str) -> None:
        """
        Set a js variable name of this instance.

        Parameters
        ----------
        variable_name : str
            Variable name to set.
        """
        from apysc._validation import string_validation
        string_validation.validate_not_empty_string(string=variable_name)
        if not hasattr(self, '_variable_name_history'):
            self._variable_name_history = []
        self._variable_name = variable_name
        self._variable_name_history.append(variable_name)

    def _get_previous_variable_name(self) -> str:
        """
        Get a previous variable name.

        Returns
        -------
        previous_variable_name : str
            A previous-variable name of this instance.
            If that value is not existing, this interface
            returns a blank string.
        """
        if not hasattr(self, '_variable_name_history'):
            return ''
        if len(self._variable_name_history) <= 1:
            return ''
        return self._variable_name_history[-2]



"""Class implementation for the scale-y from the given point
animation value.
"""

from typing import Generic
from typing import TypeVar
from typing import Union

from apysc._animation.animation_base import AnimationBase
from apysc._animation.easing import Easing
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.variable_name_interface import VariableNameInterface

_T = TypeVar('_T', bound=VariableNameInterface)


class AnimationScaleYFromPoint(AnimationBase[_T], Generic[_T]):
    """
    The animation class for a scale-y from the given point.

    References
    ----------
    - animation_scale_from_point interfaces document
        - https://simon-ritchie.github.io/apysc/animation_scale_x_and_y_from_point.html  # noqa
    - Animation interfaces duration setting document
        - https://simon-ritchie.github.io/apysc/animation_duration.html
    - Animation interfaces delay setting document
        - https://simon-ritchie.github.io/apysc/animation_delay.html
    - Each animation interface return value document
        - https://simon-ritchie.github.io/apysc/animation_return_value.html  # noqa
    - Sequential animation setting document
        - https://simon-ritchie.github.io/apysc/sequential_animation.html
    - animation_parallel interface document
        - https://simon-ritchie.github.io/apysc/animation_parallel.html
    - Easing enum document
        - https://simon-ritchie.github.io/apysc/easing_enum.html

    Examples
    --------
    >>> import apysc as ap
    >>> stage: ap.Stage = ap.Stage()
    >>> sprite: ap.Sprite = ap.Sprite()
    >>> sprite.graphics.begin_fill(color='#0af')
    >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
    ...     x=50, y=50, width=50, height=50)
    >>> animation: ap.AnimationScaleYFromPoint
    >>> animation = rectangle.animation_scale_y_from_point(
    ...     scale_y_from_point=0.5,
    ...     y=ap.Int(100),
    ...     duration=1500,
    ...     easing=ap.Easing.EASE_OUT_QUINT,
    ... )
    >>> _ = animation.start()
    """

    _scale_y_from_point: Number
    _y: Int
    _before_scale_y_from_point: Number
    _scale_y_from_point_diff_ratio: Number

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='AnimationScaleYFromPoint')
    def __init__(
            self,
            *,
            target: _T,
            scale_y_from_point: Union[float, Number],
            y: Union[int, Int],
            duration: Union[int, Int] = 3000,
            delay: Union[int, Int] = 0,
            easing: Easing = Easing.LINEAR) -> None:
        """
        The animation class for a scale-y from the given point.

        Parameters
        ----------
        target : ScaleYFromPointInterface
            A target instance of the animation target
            (e.g., `Rectangle` instance).
        scale_y_from_point : float or Number
            The final scale-y from the given point of the animation.
        y : Int or int
            Y-coordinate.
        duration : Int or int, default 3000
            Milliseconds before an animation ends.
        delay : Int or int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.

        Raises
        ------
        TypeError
            If a specified target is not a ScaleXFromPointInterface
            instance.
        """
        import apysc as ap
        from apysc._converter import to_apysc_val_from_builtin
        from apysc._display.scale_y_from_point_interface import \
            ScaleYFromPointInterface
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        variable_name: str = expression_variables_util.\
            get_next_variable_name(
                type_name=var_names.ANIMATION_SCALE_Y_FROM_POINT)
        self._y = to_apysc_val_from_builtin.\
            get_copied_int_from_builtin_val(integer=y)
        target_: VariableNameInterface = target
        if isinstance(target_, ScaleYFromPointInterface):
            target_._initialize_scale_y_from_point_if_not_initialized()
            self._before_scale_y_from_point = target_.\
                get_scale_y_from_point(y=self._y)
        else:
            raise TypeError(
                'Specified `target` argument is not a '
                f'ScaleYFromPointInterface: {type(target_)}')
        self._scale_y_from_point = to_apysc_val_from_builtin.\
            get_copied_number_from_builtin_val(
                float_or_num=scale_y_from_point)
        one: ap.Number = ap.Number(1.0)
        self._scale_y_from_point_diff_ratio = (
            one / self._before_scale_y_from_point
            * self._scale_y_from_point)
        self._set_basic_animation_settings(
            target=target,
            duration=duration,
            delay=delay,
            easing=easing)
        super(AnimationScaleYFromPoint, self).__init__(
            variable_name=variable_name)

    def _get_animation_func_expression(self) -> str:
        """
        Get a animation function expression.

        Returns
        -------
        expression : str
            Animation function expression.
        """
        from apysc._type import value_util
        diff_ratio_str: str = value_util.get_value_str_for_expression(
            value=self._scale_y_from_point_diff_ratio)
        y_str: str = value_util.get_value_str_for_expression(value=self._y)
        return f'\n  .scale(1, {diff_ratio_str}, 0, {y_str});'

    def _get_complete_event_in_handler_head_expression(self) -> str:
        """
        Get an expression to insert into the heading of the complete
        event handler.

        Returns
        -------
        expression : str
            An expression to insert into the heading of the complete
            event handler.
        """
        from apysc._display import scale_interface_helper
        from apysc._display.scale_y_from_point_interface import \
            ScaleYFromPointInterface
        from apysc._type import value_util
        expression: str = ''
        if isinstance(self._target, ScaleYFromPointInterface):
            self._target._initialize_scale_y_from_point_if_not_initialized()
            key_exp_str: str = scale_interface_helper.\
                get_coordinate_key_for_expression(coordinate=self._y).value
            target_scale_y_str: str = value_util.get_value_str_for_expression(
                value=self._target._scale_y_from_point)
            scale_y_str: str = value_util.get_value_str_for_expression(
                value=self._scale_y_from_point)
            expression = (
                f'{target_scale_y_str}[{key_exp_str}] = '
                f'{scale_y_str};'
            )
        return expression



"""This module is for the translation mapping data of the
following document:

Document file: animation_base_target.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# AnimationBase class target property interface':
    '',

    'This page explains the `AnimationBase` class `target` property interface.':  # noqa
    '',

    '## What property is this?':
    '',

    'The `target` property returns the animation target instance (e.g., `Sprite`\\, `Rectangle`).':  # noqa
    '',

    '## Basic usage':
    '',

    'Each subclass of the `AnimationBase` (e.g., `AnimationMove`\\, `AnimationX`) has the `target` getter property.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    stage_width=200, stage_height=150,\n    background_color=\'#333\', stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#00aaff\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nanimation_x: ap.AnimationX = rectangle.animation_x(x=100)\nassert isinstance(animation_x.target, ap.Rectangle)\n```':  # noqa
    '',

    '':
    '',

    '## Generic type annotation setting':
    '',

    'The `AnimationBase` class and its subclasses can set a generic type annotation. For example, the `target` property type becomes its type if you set it.\n\nThe following code sets the `[ap.Rectangle]` generic type annotation:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    stage_width=200, stage_height=150,\n    background_color=\'#333\', stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#00aaff\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nanimation_x: ap.AnimationX[ap.Rectangle] = rectangle.animation_x(x=100)\nassert isinstance(animation_x.target, ap.Rectangle)\n```':  # noqa
    '',

    'It is also sometimes useful to annotate generic type to the handler\'s `AnimationEvent`\\. This generic type annotation also affects the `target` type (`e.this.target`), as follows:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\n\ndef on_animation_complete(\n        e: ap.AnimationEvent[ap.Rectangle], options: dict) -> None:\n    """\n    The handler the animation calls when its end.\n\n    Parameters\n    ----------\n    e : ap.AnimationEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = e.this.target\n    rectangle.animation_x(x=50).start()\n\n\nap.Stage(\n    stage_width=200, stage_height=150,\n    background_color=\'#333\', stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#00aaff\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nanimation_x: ap.AnimationX = rectangle.animation_x(x=100)\nanimation_x.animation_complete(on_animation_complete)\n```':  # noqa
    '',

    '':
    '',

    '## target property API':
    '',

    '<!-- Docstring: apysc._animation.animation_base.AnimationBase.target -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface summary]** Get an animation target instance.<hr>\n\n**[Returns]**\n\n- `target`: VariableNameInterface\n  - An animation target instance.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> def on_animation_complete(\n...         e: ap.AnimationEvent[ap.Rectangle],\n...         options: dict) -> None:\n...     rectangle: ap.Rectangle = e.this.target\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> _ = rectangle.animation_x(\n...     x=100,\n... ).animation_complete(on_animation_complete).start()\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: graphics_base_flip_interfaces.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# GraphicsBase flip_x and flip_y interfaces':
    '',

    'This page explains the `GraphicsBase` class (base class of each graphics, such as the `Rectangle`) `flip_x` and `flip_y` property interfaces.':  # noqa
    '',

    '## What interfaces are these?':
    '',

    'The `flip_x` property flips an object in the x-axis direction, and the `flip_y` property flips in the y-axis direction.':  # noqa
    '',

    '## Basic usage':
    '',

    'The `flip_x` and `flip_y` can be set a `Boolean` value. If you set the `True`\\, an object becomes flipped. Conversely, if you set the `False`\\, an object resets flipping.\n\nThe getter interface returns a `Boolean` value of a current flipping value.\n\nThe following example flips the triangle polygon in the x-axis direction and resets per second:':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass _PolygonOptions(TypedDict):\n    polygon: ap.Polygon\n\n\ndef on_timer(e: ap.TimerEvent, options: _PolygonOptions) -> None:\n    """\n    The handler that the timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    polygon: ap.Polygon = options[\'polygon\']\n    flip_x: ap.Boolean = polygon.flip_x\n    flip_x = flip_x.not_\n    polygon.flip_x = flip_x\n\n\nap.Stage(\n    stage_width=150, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\n\npolygon: ap.Polygon = sprite.graphics.draw_polygon(\n    points=[\n        ap.Point2D(x=50, y=50),\n        ap.Point2D(x=50, y=100),\n        ap.Point2D(x=100, y=75),\n    ])\noptions: _PolygonOptions = {\'polygon\': polygon}\ntimer: ap.Timer = ap.Timer(\n    on_timer, delay=1000, options=options)\ntimer.start()\n\nap.save_overall_html(\n    dest_dir_path=\'graphics_base_flip_x_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/graphics_base_flip_x_basic_usage/index.html" width="150" height="150"></iframe>\n\nThe `flip_y` interface behaves the same as the `flip_x` interface, except the axis direction.':  # noqa
    '',

    '## flip_x property API':
    '',

    '<!-- Docstring: apysc._display.flip_x_interface.FlipXInterface.flip_x -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface summary]** Get a boolean value whether the x-axis is flipping or not.<hr>\n\n**[Returns]**\n\n- `flip_x`: Boolean\n  - A boolean value whether the x-axis is flipping or not.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> polygon: ap.Polygon = sprite.graphics.draw_polygon(\n...     points=[\n...         ap.Point2D(x=0, y=0),\n...         ap.Point2D(x=0, y=50),\n...         ap.Point2D(x=50, y=25),\n...     ])\n>>> polygon.flip_x = ap.Boolean(True)\n>>> polygon.flip_x\nBoolean(True)\n```':  # noqa
    '',

    '':
    '',

    '## flip_y property API':
    '',

    '<!-- Docstring: apysc._display.flip_y_interface.FlipYInterface.flip_y -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface summary]** Get a boolean value whether the y-axis is flipping or not.<hr>\n\n**[Returns]**\n\n- `flip_y`: Boolean\n  - A boolean value whether the y-axis is flipping or not.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> polygon: ap.Polygon = sprite.graphics.draw_polygon(\n...     points=[\n...         ap.Point2D(x=0, y=0),\n...         ap.Point2D(x=50, y=0),\n...         ap.Point2D(x=25, y=50),\n...     ])\n>>> polygon.flip_y = ap.Boolean(True)\n>>> polygon.flip_y\nBoolean(True)\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: import_conventions.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Import conventions':
    '',

    'The apysc library recommends using the `import apysc as ap` to import the package, like the `import numpy as np`\\, `import pandas as pd`\\, `import tkinter as tk`\\, or something like that.\n\nThe root package module packages all interfaces you need (for instance, `Sprite`\\, `Int`\\, `Stage`).\n\nAlso, it packages internal logic modules with the underscore prefix like the `_file`\\. These packages are not necessary for you.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nnumber_1: ap.Number = int_1 + ap.Number(10.5)\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: animation_base_start.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# AnimationBase class start interface':
    '',

    'This page explains the `AnimationBase` class `start` method interface.':  # noqa
    '',

    '## What interface is this?':
    '',

    'The `start` method interface starts the target animation. This interface returns the `AnimationBase` subclass instance is when you call each animation interface, such as the `animation_move` or `animation_x`\\, and it has the `start` method.':  # noqa
    '',

    '## Basic usage':
    '',

    'Notes: you need to call the `start` method to start an animation after the calling of the animation method, such as the `animation_x`\\, as follows:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    stage_width=200, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\nanimation_x: ap.AnimationX = rectangle.animation_x(\n    x=100, duration=3000, delay=3000)\nanimation_x.start()\n\nap.save_overall_html(\n    dest_dir_path=\'./animation_base_start_basic_usage_1/\')\n```':  # noqa
    '',

    '<iframe src="static/animation_base_start_basic_usage_1/index.html" width="200" height="150"></iframe>\n\nYou can also use the method chain for code simplicity:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    stage_width=200, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\nrectangle.animation_x(x=100, duration=3000, delay=3000).start()\n\nap.save_overall_html(\n    dest_dir_path=\'./animation_base_start_basic_usage_2/\')\n```':  # noqa
    '',

    '<iframe src="static/animation_base_start_basic_usage_2/index.html" width="200" height="150"></iframe>':  # noqa
    '',

    '## start API':
    '',

    '<!-- Docstring: apysc._animation.animation_base.AnimationBase.start -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `start(self) -> \'AnimationBase\'`<hr>\n\n**[Interface summary]** Start an animation with current settings.<hr>\n\n**[Returns]**\n\n- `self`: AnimatonBase\n  - This instance.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> _ = rectangle.animation_x(x=100).start()\n```':  # noqa
    '',

}



"""This module is for the abstract base class implementation
for if, else if, and else.
"""

from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import Optional
from typing import Type

from apysc._expression.indent_num import Indent
from apysc._type.boolean import Boolean


class IfBase(ABC):

    _condition: Optional[Boolean]
    _locals: Dict[str, Any]
    _globals: Dict[str, Any]
    _snapshot_name: str
    _indent: Indent

    def __init__(
            self,
            *,
            condition: Optional[Boolean],
            locals_: Optional[Dict[str, Any]] = None,
            globals_: Optional[Dict[str, Any]] = None) -> None:
        """
        A class to append if (else if and else) branch instruction
        expression.

        Parameters
        ----------
        condition : Boolean or None
            Boolean value to be used for judgment.
        locals_ : dict or None, default None
            Current scope's local variables. Set locals() value
            to this argument. If specified, this interface
            reverts all local scope VariableNameInterface
            variables (like Int, Sprite) at the end of the
            `If` or the other branch scope. This setting is
            useful when you don't want to update each variable
            by implementing the `If` or the other scope.
        globals_ : dict or None, default None
            Current scope's global variables. Set the globals()
            value to this argument. This setting works the same
            way as the `locals_` argument.

        References
        ----------
        - If document
            - https://simon-ritchie.github.io/apysc/if.html
        - Elif document
            - https://simon-ritchie.github.io/apysc/elif.html
        - Else document
            - https://simon-ritchie.github.io/apysc/else.html
        - Each branch instruction class's scope variables reverting setting
            - https://simon-ritchie.github.io/apysc/branch_instruction_variables_reverting_setting.html  # noqa
        """
        if locals_ is None:
            locals_ = {}
        if globals_ is None:
            globals_ = {}
        self._condition = condition
        self._locals = locals_
        self._globals = globals_
        self._indent = Indent()

    def __enter__(self) -> Any:
        """
        Method to be called when beginning of the with statement.

        Returns
        -------
        self : IfBase
            This instance.
        """
        from apysc._type import revert_interface
        self._snapshot_name = \
            revert_interface.make_snapshots_of_each_scope_vars(
                locals_=self._locals, globals_=self._globals)
        self._append_enter_expression()
        self._indent.__enter__()
        return self

    @abstractmethod
    def _append_enter_expression(self) -> None:
        """
        Append branch instruction start expression.
        """

    def __exit__(
            self, exc_type: Type,
            exc_value: Any,
            traceback: Any) -> None:
        """
        Method to be called when ending of the with statement.

        Parameters
        ----------
        exc_type : Type
            Exception type.
        exc_value : *
            Exception value.
        traceback : *
            Traceback value.
        """
        from apysc._type import revert_interface
        revert_interface.revert_each_scope_vars(
            snapshot_name=self._snapshot_name,
            locals_=self._locals, globals_=self._globals)
        self._indent.__exit__()
        self._append_exit_expression()
        self._set_last_scope()

    def _append_exit_expression(self) -> None:
        """
        Append if branch instruction ending expression.
        """
        import apysc as ap
        ap.append_js_expression(expression='}')

    @abstractmethod
    def _set_last_scope(self) -> None:
        """
        Set expression last scope value.
        """

    def _last_scope_is_if_or_elif(self) -> bool:
        """
        Get a boolean value whether the last scope is If or Elif.

        Returns
        -------
        result : bool
            If the last scope is If or Else, this interface
            returns True.
        """
        from apysc._expression import last_scope
        from apysc._expression.last_scope import LastScope
        last_scope_: LastScope = last_scope.get_last_scope()
        if last_scope_ != LastScope.IF and last_scope_ != LastScope.ELIF:
            return False
        return True



"""This module is for the translation mapping data of the
following document:

Document file: array_remove_and_remove_at.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Array class remove and remove_at interfaces':
    '',

    'This page explains the `Array` class `remove` and `remove_at` method interfaces.':  # noqa
    '',

    '## What interfaces are these?':
    '',

    'The `remove` method removes a specified value from an array, and the `remove_at` method removes a specified index value.':  # noqa
    '',

    '## Basic usage':
    '',

    'The `remove` method requires target value at the first argument, as follows:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\narr: ap.Array[int] = ap.Array([1, 2, 3])\narr.remove(value=2)\nassert arr == [1, 3]\n```':  # noqa
    '',

    'The `remove_at` method requires index (`int` or `Int` value) at the first argument, as follows:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\narr: ap.Array[int] = ap.Array([1, 2, 3])\narr.remove_at(index=1)\nassert arr == [1, 3]\n```':  # noqa
    '',

    '':
    '',

    '## remove API':
    '',

    '<!-- Docstring: apysc._type.array.Array.remove -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `remove(self, value:~T) -> None`<hr>\n\n**[Interface summary]** Remove a specified value from this array.<hr>\n\n**[Parameters]**\n\n- `value`: Any\n  - Value to remove.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> arr: ap.Array = ap.Array([1, 3, 5])\n>>> arr.remove(3)\n>>> arr\nArray([1, 5])\n```':  # noqa
    '',

    '':
    '',

    '## remove_at API':
    '',

    '<!-- Docstring: apysc._type.array.Array.remove_at -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `remove_at(self, index:Union[int, apysc._type.int.Int]) -> None`<hr>\n\n**[Interface summary]** Remove a specified index value from this array.<hr>\n\n**[Parameters]**\n\n- `index`: Int or int\n  - Index to remove value.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> arr: ap.Array = ap.Array([1, 2, 3])\n>>> arr.remove_at(1)\n>>> arr\nArray([1, 3])\n```':  # noqa
    '',

}



"""This module is the round-dot settings' class implementation
for a line.
"""

from typing import Union

from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.dictionary import Dictionary
from apysc._type.int import Int


class LineRoundDotSetting(Dictionary[str, Int]):
    """
    Round dot setting class for a line.

    References
    ----------
    - Graphics line_style interface document
        - https://simon-ritchie.github.io/apysc/graphics_line_style.html  # noqa

    Examples
    --------
    >>> import apysc as ap
    >>> stage: ap.Stage = ap.Stage()
    >>> sprite: ap.Sprite = ap.Sprite()
    >>> sprite.graphics.line_style(color='#fff', thickness=10)
    >>> line: ap.Line = sprite.graphics.draw_line(
    ...     x_start=50, y_start=50, x_end=150, y_end=50)
    >>> line.line_round_dot_setting = ap.LineRoundDotSetting(
    ...     round_size=10, space_size=5)
    >>> line.line_round_dot_setting.round_size
    Int(10)

    >>> line.line_round_dot_setting.space_size
    Int(5)
    """

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='LineRoundDotSetting')
    def __init__(
            self, round_size: Union[int, Int],
            space_size: Union[int, Int]) -> None:
        """
        Round dot setting class for line.

        Parameters
        ----------
        round_size : int or Int
            Dot round size.
        space_size : int or Int
            Blank space size between dots.

        References
        ----------
        - Graphics line_style interface document
            - https://simon-ritchie.github.io/apysc/graphics_line_style.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=10)
        >>> line: ap.Line = sprite.graphics.draw_line(
        ...     x_start=50, y_start=50, x_end=150, y_end=50)
        >>> line.line_round_dot_setting = ap.LineRoundDotSetting(
        ...     round_size=10, space_size=5)
        >>> line.line_round_dot_setting.round_size
        Int(10)

        >>> line.line_round_dot_setting.space_size
        Int(5)
        """
        import apysc as ap
        from apysc._converter.to_apysc_val_from_builtin import \
            get_copied_int_from_builtin_val
        from apysc._validation import number_validation
        number_validation.validate_nums_are_int_and_gt_zero(
            nums=[round_size, space_size])
        round_size_: ap.Int = get_copied_int_from_builtin_val(
            integer=round_size)
        space_size_: ap.Int = get_copied_int_from_builtin_val(
            integer=space_size)
        super(LineRoundDotSetting, self).__init__({
            'round_size': round_size_,
            'space_size': space_size_,
        })

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='LineRoundDotSetting')
    def round_size(self) -> Int:
        """
        Get a round size setting.

        Returns
        -------
        round_size : Int
            Round size setting.

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=10)
        >>> line: ap.Line = sprite.graphics.draw_line(
        ...     x_start=50, y_start=50, x_end=150, y_end=50)
        >>> line.line_round_dot_setting = ap.LineRoundDotSetting(
        ...     round_size=10, space_size=5)
        >>> line.line_round_dot_setting.round_size
        Int(10)
        """
        return self['round_size']

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='LineRoundDotSetting')
    def space_size(self) -> Int:
        """
        Get a space size setting.

        Returns
        -------
        space_size : Int
            Blank space size between dots.

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=10)
        >>> line: ap.Line = sprite.graphics.draw_line(
        ...     x_start=50, y_start=50, x_end=150, y_end=50)
        >>> line.line_round_dot_setting = ap.LineRoundDotSetting(
        ...     round_size=10, space_size=5)
        >>> line.line_round_dot_setting.space_size
        Int(5)
        """
        return self['space_size']



"""Class implementation for radius value interface.
"""

from typing import Dict
from typing import Union

from apysc._animation.animation_radius_interface import \
    AnimationRadiusInterface
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface


class RadiusInterface(
        AnimationRadiusInterface, RevertInterface, AttrLinkingInterface):

    _radius: Int

    def _initialize_radius_if_not_initialized(self) -> None:
        """
        Initialize _radius attribute if this interface does not
        initialize it yet.
        """
        if hasattr(self, '_radius'):
            return
        self._radius = Int(0)

        self._append_raidus_attr_linking_setting()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='RadiusInterface')
    def _append_raidus_attr_linking_setting(self) -> None:
        """
        Append a radius attribute linking settings.
        """
        self._append_applying_new_attr_val_exp(
            new_attr=self._radius, attr_name='radius')
        self._append_attr_to_linking_stack(
            attr=self._radius, attr_name='radius')

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='RadiusInterface')
    def radius(self) -> Int:
        """
        Get radius value.

        Returns
        -------
        radius : Int
            Radius value.

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> circle: ap.Circle = sprite.graphics.draw_circle(
        ...     x=100, y=100, radius=50)
        >>> circle.radius = ap.Int(75)
        >>> circle.radius
        Int(75)
        """
        from apysc._type import value_util
        self._initialize_radius_if_not_initialized()
        return value_util.get_copy(value=self._radius)

    @radius.setter
    def radius(self, value: Int) -> None:
        """
        Update radius value.

        Parameters
        ----------
        value : int or Int
            Radius value.
        """
        from apysc._html.debug_mode import DebugInfo
        with DebugInfo(
                callable_='points', args=[value], kwargs={},
                module_name=__name__,
                class_name=RadiusInterface.__name__):
            from apysc._validation import number_validation
            number_validation.validate_integer(integer=value)
            value = self._get_converted_radius_int(radius=value)
            self._radius = value
            self._radius._append_incremental_calc_substitution_expression()
            self._append_radius_update_expression()

            self._append_raidus_attr_linking_setting()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='RadiusInterface')
    def _get_converted_radius_int(
            self, *, radius: Union[int, Int]) -> Int:
        """
        Get a radius converted Int instance.

        Parameters
        ----------
        radius : int or Int
            Radius value.

        Returns
        -------
        radius : Int
            Type converted radius value.
        """
        import apysc as ap
        if not isinstance(radius, ap.Int):
            return ap.Int(radius)
        return radius

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='RadiusInterface')
    def _append_radius_update_expression(self) -> None:
        """
        Append radius value updating expression.
        """
        import apysc as ap
        from apysc._type import value_util
        self._initialize_radius_if_not_initialized()
        value_str: str = value_util.get_value_str_for_expression(
            value=self._radius)
        expression: str = (
            f'{self.variable_name}.radius({value_str});'
        )
        ap.append_js_expression(expression=expression)

    _radius_snapshots: Dict[str, int]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_radius_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_radius_snapshots',
            value=int(self._radius._value), snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._radius._value = self._radius_snapshots[snapshot_name]



"""Class implementation for graphic's base class.
"""

from typing import Union

from apysc._display import graphics
from apysc._display.display_object import DisplayObject
from apysc._display.flip_x_interface import FlipXInterface
from apysc._display.flip_y_interface import FlipYInterface
from apysc._display.rotation_around_center_interface import \
    RotationAroundCenterInterface
from apysc._display.rotation_around_point_interface import \
    RotationAroundPointInterface
from apysc._display.scale_x_from_center_interface import \
    ScaleXFromCenterInterface
from apysc._display.scale_x_from_point_interface import \
    ScaleXFromPointInterface
from apysc._display.scale_y_from_center_interface import \
    ScaleYFromCenterInterface
from apysc._display.scale_y_from_point_interface import \
    ScaleYFromPointInterface
from apysc._display.skew_x_interface import SkewXInterface
from apysc._display.skew_y_interface import SkewYInterface
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.int import Int


class GraphicsBase(
        DisplayObject, RotationAroundCenterInterface,
        RotationAroundPointInterface, ScaleXFromCenterInterface,
        ScaleYFromCenterInterface, ScaleXFromPointInterface,
        ScaleYFromPointInterface, FlipXInterface, FlipYInterface,
        SkewXInterface, SkewYInterface):

    _variable_name: str

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='GraphicsBase')
    def __init__(
            self,
            *,
            parent: 'graphics.Graphics',
            x: Union[int, Int],
            y: Union[int, Int],
            variable_name: str) -> None:
        """
        Vector graphic base class.

        Parameters
        ----------
        parent : Graphics
            Parent `Graphics` instance.
        x : int or Int
            X position.
        y : int or Int
            Y position.
        variable_name : str
            Variable name of this instance. This will be used to
            js expression.
        """
        import apysc as ap
        from apysc._display.graphics import Graphics
        from apysc._validation import display_validation
        from apysc._validation import number_validation
        from apysc._validation import string_validation

        display_validation.validate_graphics(graphics=parent)
        self.parent_graphics: Graphics = parent
        number_validation.validate_integer(integer=x)
        number_validation.validate_integer(integer=y)
        if isinstance(x, ap.Int):
            x_: ap.Int = x
        else:
            x_ = ap.Int(x)
        self._x = x_
        if isinstance(y, ap.Int):
            y_: ap.Int = y
        else:
            y_ = ap.Int(y)
        self._y = y_
        string_validation.validate_not_empty_string(string=variable_name)
        super(GraphicsBase, self).__init__(
            variable_name=variable_name)



"""This module is for the translation mapping data of the
following document:

Document file: assert_dicts_equal_and_dicts_not_equal.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# assert_dicts_equal and assert_dicts_not_equal interfaces':
    '',

    'This page explains the `assert_dicts_equal` and `assert_dicts_not_equal` function interfaces.':  # noqa
    '',

    '## What interfaces are these?':
    '',

    'The `assert_dicts_equal` function interface asserts specified two dictionaries (`Dictionary` type value) are equal. Conversely, the `assert_dicts_not_equal` function interface asserts specified two dictionaries are not equal.':  # noqa
    '',

    '## See also':
    '',

    '- [JavaScript assertion interface basic behavior](assertion_basic_behavior.md)':  # noqa
    '',

    '## Basic usage':
    '',

    'Both of the `assert_dicts_equal` and `assert_dicts_not_equal` interfaces require the `left` and `right` arguments. The `msg` argument is optional.\n\nYou can specify a Python built-in `dict` and `Dictionary` values to these arguments.\n\nThe following example (`assert_dicts_equal` and values are equal) passes:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nstage: ap.Stage = ap.Stage(\n    stage_width=0, stage_height=0, background_color=\'#333\',\n    stage_elem_id=\'stage\')\n\ndict_val: ap.Dictionary = ap.Dictionary({\'a\': 10, \'b\': 20})\nap.assert_dicts_equal(\n    left={\'a\': 10, \'b\': 20}, right=dict_val,\n    msg=\'Values are not equal!\')\n\nap.save_overall_html(\n    dest_dir_path=\'assert_dicts_equal_basic_usage_1/\')\n```':  # noqa
    '',

    '':
    '',

    '```\n[assert_dicts_equal]\nLeft value: {a: 10, b: 20} right value: dct_1\n```':  # noqa
    '',

    '<iframe src="static/assert_dicts_equal_basic_usage_1/index.html" width="0" height="0"></iframe>\n\nThe following example (`assert_dicts_equal` and values are not equal)  fails:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nstage: ap.Stage = ap.Stage(\n    stage_width=0, stage_height=0, background_color=\'#333\',\n    stage_elem_id=\'stage\')\n\ndict_val: ap.Dictionary = ap.Dictionary({\'a\': 10, \'b\': 20})\nap.assert_dicts_equal(\n    left={\'a\': 30}, right=dict_val, msg=\'Values are not equal!\')\n\nap.save_overall_html(\n    dest_dir_path=\'assert_dicts_equal_basic_usage_2/\')\n```':  # noqa
    '',

    '':
    '',

    '```\n[assert_dicts_equal]\nLeft value: {a: 30} right value: dct_1\n...\nAssertion failed: Values are not equal!\n```':  # noqa
    '',

    '<iframe src="static/assert_dicts_equal_basic_usage_2/index.html" width="0" height="0"></iframe>\n\nThe following example (`assert_dicts_not_equal` and values are not equal) passes:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nstage: ap.Stage = ap.Stage(\n    stage_width=0, stage_height=0, background_color=\'#333\',\n    stage_elem_id=\'stage\')\n\ndict_val: ap.Dictionary = ap.Dictionary({\'a\': 10, \'b\': 20})\nap.assert_dicts_not_equal(\n    left={\'a\': 30}, right=dict_val, msg=\'Values are equal!\')\n\nap.save_overall_html(\n    dest_dir_path=\'assert_dicts_not_equal_basic_usage_1/\')\n```':  # noqa
    '',

    '':
    '',

    '```\n[assert_dicts_not_equal]\nLeft value: {a: 30} right value: dct_1\n```':  # noqa
    '',

    '<iframe src="static/assert_dicts_not_equal_basic_usage_1/index.html" width="0" height="0"></iframe>':  # noqa
    '',

    '## Notes for the assert_equal and assert_not_equal interfaces':
    '',

    'If a `Dictionary` value is specified to the `assert_equal` or `assert_not_equal` interface\'s argument value, then the `assert_dicts_equal` or `assert_dicts_not_equal` interfaces will be called instead of the `assert_equal` or `assert_not_equal` interfaces automatically.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nstage: ap.Stage = ap.Stage(\n    stage_width=0, stage_height=0, background_color=\'#333\',\n    stage_elem_id=\'stage\')\n\ndict_val: ap.Dictionary = ap.Dictionary({\'a\': 30})\nap.assert_equal(\n    left={\'a\': 30}, right=dict_val,\n    msg=\'Values are not equal!\')\n\nap.save_overall_html(\n    dest_dir_path=\'assert_dicts_equal_notes_for_assert_equal/\')\n```':  # noqa
    '',

    '':
    '',

    '```\n[assert_dicts_equal]\nLeft value: {a: 30} right value: dct_1\n```':  # noqa
    '',

    '<iframe src="static/assert_dicts_equal_notes_for_assert_equal/index.html" width="0" height="0"></iframe>':  # noqa
    '',

    '## assert_dicts_equal API':
    '',

    '<!-- Docstring: apysc._console.assertion.assert_dicts_equal -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `assert_dicts_equal(left:Any, right:Any, *, msg:str=\'\') -> None`<hr>\n\n**[Interface summary]** JavaScript assertion interface for Dictionary values equal condition.<hr>\n\n**[Parameters]**\n\n- `left`: *\n  - Left-side value to compare.\n- `right`: *\n  - Right-side value to compare.\n- `msg`: str, optional\n  - Message to display when assertion failed.\n\n<hr>\n\n**[Notes]**\n\nThis interface is used instead of assert_equal for Dictionary class comparison (JavaScript can not compare dictionary (Object) directly, like a Python, for example, `{"a": 10} === {"a": 10}` becomes false).<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> dict_1: ap.Dictionary = ap.Dictionary({\'a\': 10})\n>>> dict_2: ap.Dictionary = ap.Dictionary({\'a\': 10})\n>>> ap.assert_dicts_equal(dict_1, dict_2)\n```':  # noqa
    '',

    '':
    '',

    '## assert_dicts_not_equal API':
    '',

    '<!-- Docstring: apysc._console.assertion.assert_dicts_not_equal -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `assert_dicts_not_equal(left:Any, right:Any, *, msg:str=\'\') -> None`<hr>\n\n**[Interface summary]** JavaScript assertion interface for Dictionary values not equal condition.<hr>\n\n**[Parameters]**\n\n- `left`: *\n  - Left-side value to compare.\n- `right`: *\n  - Right-side value to compare.\n- `msg`: str, optional\n  - Message to display when assertion failed.\n\n<hr>\n\n**[Notes]**\n\nThis interface is used instead of assert_not_equal for Dictionary class comparison (JavaScript can not compare dictionary (Object) directly, like a Python, for example, `{"a": 10} !== {"a": 10}` becomes true).<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> dict_1: ap.Dictionary = ap.Dictionary({\'a\': 10})\n>>> dict_2: ap.Dictionary = ap.Dictionary({\'a\': 20})\n>>> ap.assert_dicts_not_equal(dict_1, dict_2)\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: animation_complete.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# AnimationBase class animation_complete interface':
    '',

    'This page explains the `AnimationBase` class `animation_complete` method interface.':  # noqa
    '',

    '## What interface is this?':
    '',

    'The `animation_complete` method binds a handler that the animation calls when its end.\n\nThe handler\'s arguments require the event instance (`ap.AnimationEvent`) at the first argument and the options dictionary at the second argument.':  # noqa
    '',

    '## Basic usage':
    '',

    'The `animation_complete` method requires a handler at the first argument and the optional options dictionary at the second argument.\n\nThe following example calls the `animation_complete` method at the x-coordinate animation end. It starts another animation to reset the x-coordinate:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nDURATION: int = 1000\n\n\ndef on_animation_complete_1(\n        e: ap.AnimationEvent[ap.Rectangle],\n        options: dict) -> None:\n    """\n    The handler that animation calls when its end.\n\n    Parameters\n    ----------\n    e : ap.AnimationEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = e.this.target\n    animation_x: ap.AnimationX = rectangle.animation_x(\n        x=50, duration=DURATION)\n    animation_x.animation_complete(on_animation_complete_2)\n    animation_x.start()\n\n\ndef on_animation_complete_2(\n        e: ap.AnimationEvent[ap.Rectangle], options: dict) -> None:\n    """\n    The handler that animation calls when its end.\n\n    Parameters\n    ----------\n    e : ap.AnimationEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = e.this.target\n    animation_x: ap.AnimationX = rectangle.animation_x(\n        x=100, duration=DURATION)\n    animation_x.animation_complete(on_animation_complete_1)\n    animation_x.start()\n\n\nap.Stage(\n    stage_width=200, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\nanimation_x: ap.AnimationX = rectangle.animation_x(\n    x=100, duration=1000)\nanimation_x.animation_complete(on_animation_complete_1)\nanimation_x.start()\n\nap.save_overall_html(\n    dest_dir_path=\'./animation_complete_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/animation_complete_basic_usage/index.html" width="200" height="150"></iframe>':  # noqa
    '',

    '## Notes about the other interface calling order':
    '',

    'You can only call the `animation_complete` before the animation start, so if you call the `animation_complete` method after the `start` method, it raises an exception:':  # noqa
    '',

    '```py\nimport apysc as ap\n\n\ndef on_animation_complete(\n        e: ap.AnimationEvent[ap.Rectangle],\n        options: dict) -> None:\n    """\n    The handler that animation calls when its end.\n\n    Parameters\n    ----------\n    e : ap.AnimationEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    ap.trace(\'Animation complete!\')\n\n\nap.Stage(\n    stage_width=200, stage_height=200, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\nanimation_move: ap.AnimationMove = rectangle.animation_move(\n    x=100, y=100, duration=1000)\nanimation_move.start()\nanimation_move.animation_complete(on_animation_complete)\n```':  # noqa
    '',

    '':
    '',

    '```\nException: This interface can not be called after the animation is started.\n```':  # noqa
    '',

    'The calling of the `animation_complete` method before the `start` method works correctly:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\n\ndef on_animation_complete(\n        e: ap.AnimationEvent[ap.Rectangle],\n        options: dict) -> None:\n    """\n    The handler that animation calls when its end.\n\n    Parameters\n    ----------\n    e : ap.AnimationEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    ap.trace(\'Animation complete!\')\n\n\nap.Stage(\n    stage_width=200, stage_height=200, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\nanimation_move: ap.AnimationMove = rectangle.animation_move(\n    x=100, y=100, duration=1000)\nanimation_move.animation_complete(on_animation_complete)\nanimation_move.start()\n```':  # noqa
    '',

    '':
    '',

    '## animation_complete API':
    '',

    '<!-- Docstring: apysc._animation.animation_base.AnimationBase.animation_complete -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `animation_complete(self, handler:Callable[[_ForwardRef(\'animation_event.AnimationEvent\'), ~_O], NoneType], *, options:Union[~_O, NoneType]=None) -> \'AnimationBase\'`<hr>\n\n**[Interface summary]** Add an animation complete event listener setting.<hr>\n\n**[Parameters]**\n\n- `handler`: _Handler\n  - A callable that an instance calls when an animation is complete.\n- `options`: dict or None, default None\n  - Optional arguments dictionary to be passed to a handler.\n\n<hr>\n\n**[Returns]**\n\n- `self`: AnimatonBase\n  - This instance.\n\n<hr>\n\n**[Raises]**\n\n- Exception: If calling this interface after an animation starts\n\n<hr>\n\n**[Notes]**\n\nThis interface can only use before an animation starts<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> def on_animation_complete(\n...         e: ap.AnimationEvent[ap.Rectangle],\n...         options: dict) -> None:\n...     ap.trace(\'Animation completed!\')\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> _ = rectangle.animation_x(\n...     x=100,\n... ).animation_complete(on_animation_complete).start()\n```':  # noqa
    '',

    '<hr>\n\n**[References]**\n\n- [About the handler options’ type document](https://simon-ritchie.github.io/apysc/about_handler_options_type.html)':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: append_js_expression.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# append_js_expression interface':
    '',

    'This page explains the `append_js_expression` function interface.':
    '',

    '## What interface is this?':
    '',

    'The `append_js_expression` function interface appends any JavaScript to the exported HTML at any position. This interface may be helpful if you need to use the apysc not supported interfaces or irregular JavaScript implementation, like the Django template tags or parameters (e.g., `{% if ... %}`).':  # noqa
    '',

    '## Basic usage':
    '',

    'The `append_js_expression` function requires JavaScript string at the argument.\n\nThe following example appends the `console.log` JavaScript calling at the rectangle click handler:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\n\ndef on_click(\n        e: ap.MouseEvent[ap.Rectangle], options: dict) -> None:\n    """\n    The handler that the animation calls when its end.\n\n    Parameters\n    ----------\n    e : ap.MouseEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    ap.append_js_expression(\n        expression=\'console.log("The rectangle is clicked!");\')\n\n\nap.Stage(\n    stage_width=150, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nrectangle.click(on_click)\n\nap.save_overall_html(\n    dest_dir_path=\'append_js_expression_basic_usage/\')\n```':  # noqa
    '',

    'If you click the following rectangle, the handler displays the `The rectangle is clicked!` message on the browser console (please press the F12 key).\n\n<iframe src="static/append_js_expression_basic_usage/index.html" width="150" height="150"></iframe>':  # noqa
    '',

    '## append_js_expression API':
    '',

    '<!-- Docstring: apysc._expression.expression_data_util.append_js_expression -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `append_js_expression(expression:str) -> None`<hr>\n\n**[Interface summary]** Append js expression.<hr>\n\n**[Parameters]**\n\n- `expression`: str\n  - JavaScript Expression string.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> ap.append_js_expression(expression=\'console.log("Hello!")\')\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: graphics_base_scale_from_center.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# GraphicsBase scale_x_from_center and scale_y_from_center interfaces':
    '',

    'This page explains the `GraphicsBase` class (base class of each graphics, such as the `Rectangle`) `scale_x_from_center` and `scale_y_from_center` property interfaces.':  # noqa
    '',

    '## What interfaces are these?':
    '',

    'The `scale_x_from_center` property changes the object\'s horizontal scale, and the `scale_y_from_center` property changes the object\'s vertical scale. These scaling interfaces change the scale from the center coordinates of each object.':  # noqa
    '',

    '## Basic usage':
    '',

    'Each property getter interface returns a `Number` value. The setter interfaces also require a `Number` to update scales (If 0.0 is specified, the object becomes invisible. 1.0 becomes the default scale, and 2.0 becomes the twice-scale value).\n\nThe following example shows the default scale rectangle (left), horizontally half-scaled rectangle (center), vertically half-scaled rectangle (right).':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    stage_width=350, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\n\nleft_rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\ncenter_rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=150, y=50, width=50, height=50)\ncenter_rectangle.scale_x_from_center = ap.Number(0.5)\nright_rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=250, y=50, width=50, height=50)\nright_rectangle.scale_y_from_center = ap.Number(0.5)\n\nap.save_overall_html(\n    dest_dir_path=\'graphics_base_scale_from_center_basic_usage_1/\')\n```':  # noqa
    '',

    '<iframe src="static/graphics_base_scale_from_center_basic_usage_1/index.html" width="350" height="150"></iframe>\n\nThese interfaces apply the scaling from the center coordinates, as follows:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    stage_width=150, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\', alpha=0.3)\n\nrectangle_1: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\nrectangle_2: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nrectangle_2.scale_x_from_center = ap.Number(0.5)\nrectangle_2.scale_y_from_center = ap.Number(0.5)\n\nrectangle_3: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nrectangle_3.scale_x_from_center = ap.Number(0.25)\nrectangle_3.scale_y_from_center = ap.Number(0.25)\n\nap.save_overall_html(\n    dest_dir_path=\'graphics_base_scale_from_center_basic_usage_2/\')\n```':  # noqa
    '',

    '<iframe src="static/graphics_base_scale_from_center_basic_usage_2/index.html" width="150" height="150"></iframe>\n\nThe `+=` and `-=` operators are also supported:':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass _RectanglesOptions(TypedDict):\n    rectangle_1: ap.Rectangle\n    rectangle_2: ap.Rectangle\n    direction: ap.Int\n\n\ndef on_timer(e: ap.TimerEvent, options: _RectanglesOptions) -> None:\n    """\n    The handler that the timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle_1: ap.Rectangle = options[\'rectangle_1\']\n    rectangle_2: ap.Rectangle = options[\'rectangle_2\']\n    direction: ap.Int = options[\'direction\']\n\n    current_scale: ap.Number = rectangle_1.scale_x_from_center\n    condition_1: ap.Boolean = current_scale >= 2.0\n    condition_2: ap.Boolean = current_scale <= 0.05\n    with ap.If(condition_1):\n        direction.value = -1\n    with ap.Elif(condition_2):\n        direction.value = 1\n\n    rectangle_1.scale_x_from_center += direction * 0.03\n    rectangle_2.scale_y_from_center += direction * 0.03\n\n\nap.Stage(\n    stage_width=150, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\n\nsprite.graphics.begin_fill(color=\'#0af\', alpha=0.5)\nrectangle_1: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\nsprite.graphics.begin_fill(color=\'#f0a\', alpha=0.5)\nrectangle_2: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\n\ndirection: ap.Int = ap.Int(1.0)\noptions: _RectanglesOptions = {\n    \'rectangle_1\': rectangle_1, \'rectangle_2\': rectangle_2,\n    \'direction\': direction}\ntimer: ap.Timer = ap.Timer(\n    on_timer, delay=ap.FPS.FPS_60,\n    options=options)\ntimer.start()\n\nap.save_overall_html(\n    dest_dir_path=\'graphics_base_scale_from_center_basic_usage_3/\')\n```':  # noqa
    '',

    '<iframe src="static/graphics_base_scale_from_center_basic_usage_3/index.html" width="150" height="150"></iframe>':  # noqa
    '',

    '## scale_x_from_center property API':
    '',

    '<!-- Docstring: apysc._display.scale_x_from_center_interface.ScaleXFromCenterInterface.scale_x_from_center -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface summary]** Get a scale-x value from the center of this instance.<hr>\n\n**[Returns]**\n\n- `scale_x_from_center`: ap.Number\n  - Scale-x value from the center of this instance.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> rectangle.scale_x_from_center = ap.Number(1.5)\n>>> rectangle.scale_x_from_center\nNumber(1.5)\n```':  # noqa
    '',

    '':
    '',

    '## scale_y_from_center property API':
    '',

    '<!-- Docstring: apysc._display.scale_y_from_center_interface.ScaleYFromCenterInterface.scale_y_from_center -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface summary]** Get a scale-y value from the center of this instance.<hr>\n\n**[Returns]**\n\n- `scale_y_from_center`: ap.Number\n  - Scale-y value from the center of this instance.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> rectangle.scale_y_from_center = ap.Number(1.5)\n>>> rectangle.scale_y_from_center\nNumber(1.5)\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: animation_delay.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Animation interfaces delay setting':
    '',

    'This page explains the animation interfaces `delay` setting.':
    '',

    '## What setting is this?':
    '',

    'The `delay` setting determines the delay time before the animation starts. For instance, if you specify 3000 to the `delay` argument, the animation starts after 2 seconds after.\n\nEach animation method interface (such as the `animation_move`\\, `animation_x`\\, and so on) has the `delay` argument.':  # noqa
    '',

    '## Basic usage':
    '',

    'The following example sets 2 seconds between each x-coordinate animation (pause 2 seconds before animation starts):':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nDURATION: int = 3000\nDELAY: int = 2000\nEASING: ap.Easing = ap.Easing.EASE_OUT_QUINT\n\n\ndef on_animation_complete_1(\n        e: ap.AnimationEvent[ap.Rectangle],\n        options: dict) -> None:\n    """\n    The handler that animation calls when its end.\n\n    Parameters\n    ----------\n    e : ap.AnimationEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = e.this.target\n    animation_x: ap.AnimationX = rectangle.animation_x(\n        x=50, duration=DURATION, delay=DELAY, easing=EASING)\n    animation_x.animation_complete(on_animation_complete_2)\n    animation_x.start()\n\n\ndef on_animation_complete_2(\n        e: ap.AnimationEvent[ap.Rectangle], options: dict) -> None:\n    """\n    The handler that animation calls when its end.\n\n    Parameters\n    ----------\n    e : ap.AnimationEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = e.this.target\n    animation_x: ap.AnimationX = rectangle.animation_x(\n        x=300, duration=DURATION, delay=DELAY, easing=EASING)\n    animation_x.animation_complete(on_animation_complete_1)\n    animation_x.start()\n\n\nap.Stage(\n    stage_width=400, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\nanimation_x: ap.AnimationX = rectangle.animation_x(\n    x=300, duration=DURATION, delay=DELAY, easing=EASING)\nanimation_x.animation_complete(on_animation_complete_1)\nanimation_x.start()\n\nap.save_overall_html(\n    dest_dir_path=\'./animation_delay_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/animation_delay_basic_usage/index.html" width="400" height="150"></iframe>':  # noqa
    '',

}



"""Class implementation for line thickness interface.
"""

from typing import Dict
from typing import Union

from apysc._animation.animation_line_thickness_interface import \
    AnimationLineThicknessInterface
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface


class LineThicknessInterface(
        AnimationLineThicknessInterface,
        RevertInterface,
        AttrLinkingInterface):

    _line_thickness: Int

    def _initialize_line_thickness_if_not_initialized(self) -> None:
        """
        Initialize _line_thickness attribute if this interface
        does not initialize it yet.
        """
        if hasattr(self, '_line_thickness'):
            return
        self._line_thickness = Int(1)

        self._append_line_thickness_attr_linking_setting()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='LineThicknessInterface')
    def _append_line_thickness_attr_linking_setting(self) -> None:
        """
        Append a line thickness attribute linking settings.
        """
        self._append_applying_new_attr_val_exp(
            new_attr=self._line_thickness, attr_name='line_thickness')
        self._append_attr_to_linking_stack(
            attr=self._line_thickness, attr_name='line_thickness')

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='LineThicknessInterface')
    def line_thickness(self) -> Int:
        """
        Get this instance's line thickness.

        Returns
        -------
        line_thickness : Int
            Current line thickness.

        References
        ----------
        - Graphics line_thickness interface document
            - https://simon-ritchie.github.io/apysc/graphics_line_thickness.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(
        ...     color='#fff', thickness=5)
        >>> line: ap.Line = sprite.graphics.draw_line(
        ...     x_start=50, y_start=50, x_end=150, y_end=50)
        >>> line.line_thickness
        Int(5)
        """
        from apysc._type import value_util
        return value_util.get_copy(value=self._line_thickness)

    @line_thickness.setter
    def line_thickness(self, value: Int) -> None:
        """
        Update this instance's line thickness.

        Parameters
        ----------
        value : Int
            Line thickness to set.

        References
        ----------
        - Graphics line_thickness interface document
            - https://simon-ritchie.github.io/apysc/graphics_line_thickness.html  # noqa
        """
        from apysc._html.debug_mode import DebugInfo
        with DebugInfo(
                callable_='line_thickness', args=[value],
                kwargs={}, module_name=__name__,
                class_name=LineThicknessInterface.__name__):
            self._update_line_thickness_and_skip_appending_exp(value=value)
            self._line_thickness.\
                _append_incremental_calc_substitution_expression()
            self._append_line_thickness_update_expression()

            self._append_line_thickness_attr_linking_setting()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='LineThicknessInterface')
    def _append_line_thickness_update_expression(self) -> None:
        """
        Append line thickness update expression.
        """
        import apysc as ap
        from apysc._type import value_util
        line_thickness_str: str = value_util.get_value_str_for_expression(
            value=self.line_thickness)
        expression: str = (
            f'{self.variable_name}.attr({{"stroke-width": '
            f'{line_thickness_str}}});'
        )
        ap.append_js_expression(expression=expression)

    def _update_line_thickness_and_skip_appending_exp(
            self, *, value: Union[int, Int]) -> None:
        """
        Update line thickness and skip appending expression.

        Parameters
        ----------
        value : Int or int
            Line thickness to set.
        """
        import apysc as ap
        from apysc._validation import number_validation
        number_validation.validate_integer(integer=value)
        number_validation.validate_num_is_gte_zero(num=value)
        if isinstance(value, ap.Int):
            value_: ap.Int = value
        else:
            value_ = Int(value)
        self._line_thickness = value_

    _line_thickness_snapshots: Dict[str, int]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_line_thickness_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_line_thickness_snapshots',
            value=self._line_thickness._value, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._line_thickness._value = self._line_thickness_snapshots[
            snapshot_name]



"""Class Implementation for an event.
"""

from typing import Generic
from typing import Optional
from typing import TypeVar

from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.variable_name_interface import VariableNameInterface

T = TypeVar('T', bound=VariableNameInterface)


class Event(Generic[T], VariableNameInterface):
    """
    Basic event class.

    Examples
    --------
    >>> import apysc as ap
    >>> def on_custom_event(
    ...         e: ap.Event[ap.Rectangle], options: dict) -> None:
    ...     rectangle: ap.Rectangle = e.this
    ...     rectangle.fill_color = ap.String('#f0a')
    >>> stage: ap.Stage = ap.Stage()
    >>> sprite: ap.Sprite = ap.Sprite()
    >>> sprite.graphics.begin_fill(color='#0af')
    >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
    ...     x=50, y=50, width=50, height=50)
    >>> e: ap.Event = ap.Event(this=rectangle)
    >>> _ = rectangle.bind_custom_event(
    ...     custom_event_type='my_custom_event',
    ...     handler=on_custom_event, e=e)
    >>> # Do something here and then trigger the custom event
    >>> rectangle.trigger_custom_event(
    ...     custom_event_type='my_custom_event')
    """

    _this: T

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Event')
    def __init__(
            self, this: T,
            *,
            type_name: Optional[str] = None) -> None:
        """
        Basic event class.

        Parameters
        ----------
        this : VariableNameInterface
            Instance that listening event (e.g., Sprite).
        type_name : str or None, default None
            Type name to set. Only specify when inheriting
            this class.

        Examples
        --------
        >>> import apysc as ap
        >>> def on_custom_event(
        ...         e: ap.Event[ap.Rectangle], options: dict) -> None:
        ...     rectangle: ap.Rectangle = e.this
        ...     rectangle.fill_color = ap.String('#f0a')
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> e: ap.Event = ap.Event(this=rectangle)
        >>> _ = rectangle.bind_custom_event(
        ...     custom_event_type='my_custom_event',
        ...     handler=on_custom_event, e=e)
        >>> # Do something here and then trigger the custom event
        >>> rectangle.trigger_custom_event(
        ...     custom_event_type='my_custom_event')
        """
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        self._validate_type_name_and_self_type(type_name=type_name)
        self._this = this
        if type_name is None:
            type_name = var_names.EVENT
        self.variable_name = expression_variables_util.\
            get_next_variable_name(type_name=type_name)

    def _validate_type_name_and_self_type(
            self, *, type_name: Optional[str]) -> None:
        """
        Validate type_name argument is None when a self
        instance is not Event subclass, and the same is true
        for the opposite pattern.

        Parameters
        ----------
        type_name : str or None
            Type name to set.

        Raises
        ------
        ValueError
            - If type_name is not None and self instance is Event type.
            - If type_name is None and self instance is not Event type.
        """
        from apysc._type import type_util
        if type_name is not None:
            if type_util.is_same_class_instance(class_=Event, instance=self):
                raise ValueError(
                    'type_name argument can be set only when this instance '
                    'is subclass of Event.')
            return
        if not type_util.is_same_class_instance(class_=Event, instance=self):
            raise ValueError(
                'type_name argument can\'t be set when this instance '
                'is Event (this will be used by Event subclass).')

    @property
    def this(self) -> T:
        """
        Get an instance of listening to this event.

        Returns
        -------
        this : VariableNameInterface
            Instance that listening this event.

        Examples
        --------
        >>> import apysc as ap
        >>> def on_custom_event(
        ...         e: ap.Event[ap.Rectangle], options: dict) -> None:
        ...     rectangle: ap.Rectangle = e.this
        ...     rectangle.fill_color = ap.String('#f0a')
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> e: ap.Event = ap.Event(this=rectangle)
        >>> _ = rectangle.bind_custom_event(
        ...     custom_event_type='my_custom_event',
        ...     handler=on_custom_event, e=e)
        >>> # Do something here and then trigger the custom event
        >>> rectangle.trigger_custom_event(
        ...     custom_event_type='my_custom_event')
        """
        return self._this
